import pathlib
import json
import re
from jinja2 import FileSystemLoader, Environment, select_autoescape

class CCTemplater:
    def __init__(self, template_folder=None):
        self.template_folder = template_folder if template_folder else pathlib.Path("ccutils/templates")
        self.env = Environment(
            loader=FileSystemLoader(searchpath=str(self.template_folder)),
            autoescape=select_autoescape(["jinja2"]),
            lstrip_blocks=True,
            trim_blocks=True
        )
        
    def render(self, template_name, context):
        template = self.env.get_template(template_name)
        return template.render(context=context)

if __name__ == "__main__":
    context = None
    with open(r"C:\Users\mhudec\Develop\cctemplater\cctemplater\context_skeleton.json") as f:
        context = json.load(fp=f)
    ct = CCTemplater()
    print(ct.render(template_name="base_config.jinja2", context=context))


import json
import pathlib
import logging
import sys
import re
from collections import OrderedDict

INTERFACE_FLAGS_SUBSTRING_MAP = {
    "physical": [
        "Ethernet",
        "GigE"
    ],
    "virtual": [
        "Loopback",
        "Vlan"
    ],
    "port-channel": [
        "Port-channel"
    ]
}

PHYSICAL_INTERFACE_SUBSTRINGS = [
    "Ethernet",
    "GigE"
]
VIRTUAL_INTERFACE_SUBSTRINGS = [
    "Loopback",
    "Vlan"
]

class UnsortableList(list):
    def sort(self, *args, **kwargs):
        pass


class UnsortableOrderedDict(OrderedDict):
    def items(self, *args, **kwargs):
        return UnsortableList(OrderedDict.items(self, *args, **kwargs))


def get_logger(name, verbosity=4):
    """
    """

    verbosity_map = {
        1: logging.CRITICAL,
        2: logging.ERROR,
        3: logging.WARNING,
        4: logging.INFO,
        5: logging.DEBUG
    }

    threading_formatter_string = '[%(asctime)s] [%(levelname)s]\t[%(name)s][%(threadName)s][%(module)s][%(funcName)s]\t%(message)s'
    single_formatter_string = '[%(asctime)s] [%(levelname)s]\t[%(name)s][%(module)s][%(funcName)s]\t%(message)s'

    logger = logging.getLogger(name)
    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter(single_formatter_string)
    handler.setFormatter(formatter)
    if not len(logger.handlers):
        logger.addHandler(handler)
    logger.setLevel(verbosity_map[verbosity])

    # To be removed
    # TODO: Remove
    # if verbosity == 0:
    #     logger.setLevel(logging.WARN)
    # elif verbosity == 1:  # default
    #     logger.setLevel(logging.INFO)
    # elif verbosity > 1:
    #     logger.setLevel(logging.DEBUG)

    return logger


logger = get_logger(__name__, verbosity=3)


def load_json(path):
    path = check_path(path)
    data = None
    if path:
        with path.open(mode="r") as f:
            data = json.load(f)
    return data


def split_interface_name(interface: str):
    """
    This function takes in interface string such as "GigabitEthernet0/10" and returns a list containing name and number,
    such as ["GigabitEthernet", "0/10"]

    Args:
        interface (str): Interface to perform split on

    Returns:
        list: List containing name and number of interface, such as ``["GigabitEthernet", "0/10"]``

    """
    #pattern = re.compile(pattern=r"(?P<name>^[A-z\-]+(?=\d))(?P<number>[\d+\/]+)")
    pattern = re.compile(pattern=r"(?P<name>[A-z]{2,}(?:[A-z\-])*)(?P<number>\d+(?:\/\d+)*(?:\:\d+)?(?:\.\d+)?)")
    try:
        match = re.match(pattern=pattern, string=interface)
    except TypeError as e:
        logger.error("Expected string or bytes-like object, cannot match on '{}'".format(type(interface)))
        return None
    if match:
        return [match.group("name"), match.group("number")]
    else:
        logger.error("Given interface {} did not match parsing pattern.".format(interface))
        return None


def convert_interface_name(interface: str, out: str="long"):
    """
    This function converts interface names between long and short variants.
    For example Fa0/1 -> FastEthernet0/1 or the other way around.

    :param interface:
    :param out:
    :rtype: str
    :return: Interface string
    """
    short = [
        "Eth", "Et", "Se", "Fa", "Gi", "Te", "Twe", "Fo", "Hu", "Po", "Tu", "Vl", "Lo", "pw"
    ]
    long = [
        "Ethernet", "Ethernet", "Serial", "FastEthernet", "GigabitEthernet",
        "TenGigabitEthernet", "TwentyFiveGigE", "FortyGigabitEthernet", "HundredGigE", "Port-channel", "Tunnel",
        "Vlan", "Loopback", "pseudowire"
    ]
    interface_s = split_interface_name(interface)
    if interface_s is None:
        logger.error("Cannot convert given interface.")
        return interface
    else:
        given_form = None
        index = None
        name, number = interface_s
        if name in short:
            given_form = "short"
            index = short.index(name)
        elif name in long:
            given_form = "long"
            index = long.index(name)
        else:
            given_form = "unknown"
            logger.warning("Got unknown interface name: '{}'".format(name))

        if out == "long":
            # Return full version
            if given_form == "long":
                return interface
            elif given_form == "short":
                return long[index] + number
            else:
                # Error
                return interface
        elif out == "short":
            # Return full version
            if given_form == "short":
                return interface
            elif given_form == "long":
                return short[index] + number
            else:
                # Error
                return interface
        else:
            return interface


def match_to_json(match, groups):
    """
    This function converts `re` match object to dict

    :param match: re.match object
    :param groups: list
    :rtype: dict
    :return: Dictionary with matched groups
    """
    result = {k: None for k in groups}
    for group in groups:
        try:
            result[group] = match.group(group)
        except Exception:
            pass
    return result


def remove_empty_values(entry: dict) -> dict:
    return {k: v for k, v in entry.items() if v is not None}


def strip_none(data):
    # https://stackoverflow.com/questions/20558699/python-how-recursively-remove-none-values-from-a-nested-data-structure-lists-a
    if isinstance(data, dict):
        return {k: strip_none(v) for k, v in data.items() if k is not None and v is not None}
    elif isinstance(data, list):
        return [strip_none(item) for item in data if item is not None]
    elif isinstance(data, tuple):
        return tuple(strip_none(item) for item in data if item is not None)
    elif isinstance(data, set):
        return {strip_none(item) for item in data if item is not None}
    else:
        return data


def strip_false(data):
    # https://stackoverflow.com/questions/20558699/python-how-recursively-remove-false-values-from-a-nested-data-structure-lists-a
    if isinstance(data, dict):
        return {k: strip_false(v) for k, v in data.items() if k is not False and v is not False}
    elif isinstance(data, list):
        return [strip_false(item) for item in data if item is not False]
    elif isinstance(data, tuple):
        return tuple(strip_false(item) for item in data if item is not False)
    elif isinstance(data, set):
        return {strip_false(item) for item in data if item is not False}
    else:
        return data


def value_to_bool(entry: dict, keys: list, keep_none=False) -> dict:
    if not isinstance(keys, list):
        keys = [keys]
    for key in keys:
        if key not in entry.keys():
            continue
        if entry[key] is not None:
            entry[key] = True
        else:
            if keep_none:
                continue
            else:
                entry[key] = False
    return entry


def value_to_int(entry: dict, keys: list) -> dict:
    if not isinstance(keys, list):
        keys = [keys]
    for key in keys:
        if key not in entry.keys():
            continue
        try:
            entry[key] = int(entry[key])
        except (ValueError, TypeError):
            continue
    return entry


def check_path(path, create=False):
    """

    :param path:
    :param create:
    :return:
    """
    return_path = None
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    if path.exists():
        if path.is_file():
            return_path = path
        elif path.is_dir():
            return_path = path
    else:
        if create:
            if "." not in path.name[1:]:
                # Path is a folder
                path.mkdir(parents=True)
                if path.exists():
                    return_path = path
            else:
                # Path is a file
                if path.parent.exists():
                    path.touch()
                    return_path = path
                else:
                    path.parent.mkdir()
                    if path.parent.exists():
                        path.touch()
                        return_path = path
    return return_path


def interface_sort(crange, name):
    index = crange._list.index(name)
    return crange._list.index(name)


def jprint(data, indent=2):
    print(json.dumps(obj=data, indent=indent))

def has_old_pyyaml():
    try:
        from packaging import version
    except ModuleNotFoundError:
        logger.error("Cannot determine version o PyYAML, missing 'packaging' package. Acting like PyYAML >= 5.1")
        version = None
        return False
    try:
        import yaml
    except ImportError:
        logger.error("Missing Package PyYAML")
        return False

    return version.parse(yaml.__version__) < version.parse("5.1")

def get_flags_from_interface_name(interface):
    flags = set()
    for flag, subs in INTERFACE_FLAGS_SUBSTRING_MAP.items():
        for sub in subs:
            if sub in interface:
                flags.add(flag)
    if re.search(pattern=r"\d+\.d+", string=interface):
        flags.add("subinterface")
    return list(flags)

def load_excel_sheet(file, sheet_name):
    try:
        import pandas as pd
    except ImportError:
        print("To use 'load_excel_sheet' function you need to have 'pandas' installed.")
        pd = None
    if pd:
        file = check_path(file)
        if not file:
            return None
        df = pd.read_excel(io=file, sheet_name=sheet_name)
        df = df.where((pd.notnull(df)), None)
        data = {}
        for column in df.columns:
            data[column] = list(df[column])
        return data
    else:
        return None




import pathlib
import unicodedata
import yaml
import re
from ccutils.utils.common_utils import get_logger, interface_sort
from ccutils.utils.CiscoRange import CiscoRange
from ccutils.utils.CustomAnsibleDumper import CustomAnsibleDumper
from collections import OrderedDict
from pprint import pprint


try:
    import pandas as pd
except ImportError:
    print("To use 'ExcelInventory' function you need to have 'pandas' installed.")
    pd = None

class GroupDoesNotExist(Exception):
    pass

class HostDoesNotExist(Exception):
    pass

class ObjectNotFound(object):
    pass

    def __repr__(self):
        return "[ObjectNotFound]"

    def __str__(self):
        return "[ObjectNotFound]"

    def __bool__(self):
        return False

class ExcelInventory(object):
    pass

    def __init__(self, input_file, output_dir, verbosity=4):
        self.logger = get_logger(name="ExcelInventory", verbosity=verbosity)
        self.input_file = self.check_path(path=input_file, mode="file")
        self.output_dir = self.check_path(path=output_dir, mode="directory")
        self.host_vars = {}
        self.group_vars = {}
        self.hosts = {}

    def check_path(self, path, mode):
        """
        """
        self.logger.info("Checking Path: '{}'".format(path))
        try:
            if not isinstance(path, pathlib.Path):
                path = pathlib.Path(path)
            if path.exists():
                if path.is_file() and mode == "file":
                    self.logger.info("Path: '{}' Exists: File.".format(path))
                elif path.is_file() and mode == "directory":
                    self.logger.critical("Path: '{}' Exists but is not a file!".format(path))
                    path = None
                elif not path.is_file() and mode == "directory":
                    self.logger.info("Path: '{}' Exists: Directory.".format(path))
                elif not path.is_file() and mode == "file":
                    self.logger.critical("Path: '{}' Exists but is not a directory!".format(path))
                    path = None
                else:
                    self.logger.critical("Path: '{}' Unhandled error!".format(path))
            else:
                self.logger.critical("Path: '{}' Does not exist!".format(path))
        except Exception as e:
            self.logger.critical("Could not determine valid path for '{}'. Exception: {}".format(path, repr(e)))
        finally:
            return path

    def load_excel(self, path, sheet_name, index_column=None, columns_rename=None, **kwargs):
        self.logger.info("Loading file: '{}' Sheet: '{}' as DF".format(path, sheet_name))
        df = pd.read_excel(io=path, sheet_name=sheet_name, index_col=index_column, engine="openpyxl", **kwargs)
        df = df.where(pd.notnull(df), None)
        if columns_rename is not None:
            df = df.rename(columns=columns_rename)
        return df

    def duplicates_check(self, df, columns):
        results = []
        for column_name in columns:
            duplicates = df.duplicated(subset=[column_name])
            results.append(any(duplicates))
            if results[-1]:
                self.logger.warning(
                    "Found duplicated values in column '{0}': {1}".format(column_name, df[duplicates][column_name]))
        return results

    @staticmethod
    def replace_cz_chars(line):
        line = unicodedata.normalize('NFKD', line)
        output = ''
        for c in line:
            if not unicodedata.combining(c):
                output += c
        return output

    def _finditem(self, obj, key):
        if key in obj:
            return obj
        for k, v in obj.items():
            if isinstance(v, dict):
                return self._finditem(v, key)  # added return statement
        return ObjectNotFound()

    def recursive_parent_lookup(self, key, obj):
        if key in obj:
            return obj
        for v in obj.values():
            if isinstance(v, dict):
                a = self.recursive_parent_lookup(key=key, obj=v)
                if not isinstance(a, ObjectNotFound):
                    return a
        return ObjectNotFound()

    def get_ordered_interfaces(self, host):
        """
        Return interfaces as OrderedDict

        Returns:
            (:obj:`OrderedDict`): Interface section as OrderedDict

        """
        if host not in self.host_vars.keys():
            msg = "Host '{}' does not exist".format(host)
            raise HostDoesNotExist(msg)
        if "interfaces" not in self.host_vars[host].keys():
            return OrderedDict()
        interfaces_crange = CiscoRange(list(self.host_vars[host]["interfaces"].keys()))
        ordered_interfaces = OrderedDict(sorted(self.host_vars[host]["interfaces"].items(), key=lambda x: interface_sort(crange=interfaces_crange, name=x[0])))
        return ordered_interfaces

    def dump_hosts(self, outputfile):
        self.logger.info("Storing hosts as YAML file.")
        with self.output_dir.joinpath(outputfile).open(mode="w") as f:
            yaml.dump(data=self.hosts, stream=f, Dumper=CustomAnsibleDumper)

    def dump_hostvars(self, nested=False):
        self.logger.info("Storing host_vars as YAML files.")
        if self.output_dir is not None:
            host_vars_path = self.output_dir.joinpath("host_vars")
            host_vars_path.mkdir(exist_ok=True)
        for hostname, host_vars in self.host_vars.items():
            if not nested:
                path = host_vars_path.joinpath("{}.yml".format(hostname))
                with path.open(mode="w") as f:
                    data = dict(host_vars)
                    data["interfaces"] = self.get_ordered_interfaces(host=hostname)
                    yaml_string = yaml.dump(data=data, Dumper=CustomAnsibleDumper)
                    yaml_string = re.sub("'\"(.*)\"'", '"\\1"', yaml_string)
                    f.write(yaml_string)
                    # yaml.dump(data=host_vars, stream=f, Dumper=CustomAnsibleDumper)
            else:
                host_path = host_vars_path.joinpath(hostname)
                host_path.mkdir(exist_ok=True)
                general_path = host_path.joinpath("{}.yml".format(hostname))
                interfaces_path = host_path.joinpath("interfaces.yml")
                data = dict(host_vars)
                # Interfaces Section
                interfaces_data = {"interfaces": self.get_ordered_interfaces(host=hostname)}
                with interfaces_path.open(mode="w") as f:
                    yaml_string = yaml.dump(data=interfaces_data, Dumper=CustomAnsibleDumper)
                    yaml_string = re.sub("'\"(.*)\"'", '"\\1"', yaml_string)
                    f.write(yaml_string)

                # General Section
                general_data = data
                del general_data["interfaces"]
                with general_path.open(mode="w") as f:
                    yaml_string = yaml.dump(data=general_data, Dumper=CustomAnsibleDumper)
                    yaml_string = re.sub("'\"(.*)\"'", '"\\1"', yaml_string)
                    f.write(yaml_string)

    def dump_groupvars(self, nested=False):
        self.logger.info("Storing group_vars as YAML files.")
        if self.output_dir is not None:
            group_vars_path = self.output_dir.joinpath("group_vars")
            group_vars_path.mkdir(exist_ok=True)
        for groupname, group_vars in self.group_vars.items():
            if not nested:
                path = group_vars_path.joinpath("{}.yml".format(groupname))
                with path.open(mode="w") as f:
                    yaml.dump(data=group_vars, stream=f, Dumper=CustomAnsibleDumper)
            else:
                group_path = group_vars_path.joinpath(groupname)
                group_path.mkdir(exist_ok=True)
                general_path = group_path.joinpath("{}.yml".format(groupname))
                with general_path.open(mode="w") as f:
                    yaml.dump(data=group_vars, stream=f, Dumper=CustomAnsibleDumper)




from ccutils.ccparser import BaseConfigParser, BaseConfigLine
from ccutils.utils.common_utils import get_logger, interface_sort, UnsortableOrderedDict, has_old_pyyaml
from ccutils.utils import CiscoRange
import re
import json
import pathlib
from collections import OrderedDict


class ConfigToJson:
    """

    """
    def __init__(self, config, omit_empty=False, verbosity=3):
        """

        :param config: Reference to the parent BaseConfigParser object
        :param int verbosity: Logging output level
        """
        self.config = config
        self.omit_empty = omit_empty
        self.logger = get_logger(name="CTJ", verbosity=verbosity)
        self.data = {
            "interfaces": {}
        }
        self.parse_interfaces()
        self.parse_common()

    def parse_interfaces(self):
        """

        :return:
        """
        interface_lines = list(filter(lambda x: "interface" in x.type, self.config.lines))
        self.logger.debug(msg="Loaded {} interface lines.".format(len(interface_lines)))
        for interface in interface_lines:
            port_mode = interface.port_mode
            flags = interface.flags
            self.data["interfaces"][interface.name] = {"flags": flags, "unprocessed_lines": interface.get_unprocessed(return_type="text")}

            # Get Shutdown State
            self.data["interfaces"][interface.name]["shutdown"] = interface.shutdown

            # Get Description
            if interface.description or not self.omit_empty:
                self.data["interfaces"][interface.name]["description"] = interface.description

            # Get CDP
            self.data["interfaces"][interface.name]["cdp"] = interface.cdp

            # Get Logging events
            if interface.logging_events or not self.omit_empty:
                self.data["interfaces"][interface.name]["logging_events"] = interface.logging_events

            # Get channel group
            if interface.channel_group or not self.omit_empty:
                self.data["interfaces"][interface.name]["channel_group"] = interface.channel_group
            
            # Get speed and duplex
            if interface.speed or not self.omit_empty:
                self.data["interfaces"][interface.name]["speed"] = interface.speed
            if interface.duplex or not self.omit_empty:
                self.data["interfaces"][interface.name]["duplex"] = interface.duplex

            # Get Interface MTU
            if interface.mtu or not self.omit_empty:
                self.data["interfaces"][interface.name]["mtu"] = interface.mtu

            # Get Interface MTU
            if interface.bandwidth or not self.omit_empty:
                self.data["interfaces"][interface.name]["bandwidth"] = interface.bandwidth

            # Get Interface Load-Interval
            if interface.load_interval or not self.omit_empty:
                self.data["interfaces"][interface.name]["load_interval"] = interface.load_interval

            # Get Service Policies
            if interface.service_policy != {"input": None, "output": None} or not self.omit_empty:
                self.data["interfaces"][interface.name]["service_policy"] = interface.service_policy

            # Get Service Instances
            if interface.service_instances or not self.omit_empty:
                self.data["interfaces"][interface.name]["service_instances"] = interface.service_instances

            # Get negotiation
            if interface.negotiation or not self.omit_empty:
                self.data["interfaces"][interface.name]["negotiation"] = interface.negotiation





            if port_mode == "l3":
                # Get IP addresses
                self.data["interfaces"][interface.name]["l3"] = {}
                ip_addresses = interface.ip_addresses
                if len(ip_addresses):
                    self.data["interfaces"][interface.name]["l3"]["ip_addresses"] = ip_addresses
                else:
                    self.data["interfaces"][interface.name]["l3"]["ip_addresses"] = []
                # Get VRF
                if interface.vrf or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["vrf"] = interface.vrf
                # TODO: Remove "ospf_priority"
                # Get OSPF Priority
                # self.data["interfaces"][interface.name]["l3"]["ospf_priority"] = interface.ospf_priority
                # Get standby
                if interface.standby or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["standby"] = interface.standby
                    if self.data["interfaces"][interface.name]["l3"]["standby"] is not None:
                        self.data["interfaces"][interface.name]["flags"].append("standby")
                # Get Helper Address
                if interface.helper_address or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["helper_addresses"] = interface.helper_address
                ip_mtu = interface.ip_mtu
                tcp_mss = interface.tcp_mss
                if ip_mtu or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["ip_mtu"] = ip_mtu
                if tcp_mss or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["tcp_mss"] = tcp_mss
                if interface.encapsulation or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["encapsulation"] = interface.encapsulation
                if interface.ospf or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["ospf"] = interface.ospf
                if interface.isis or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["isis"] = interface.isis
                # IP Unnumbered
                if interface.ip_unnumbered_interface or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l3"]["unnumbered"] = interface.ip_unnumbered_interface

            elif port_mode == "l2":

                self.data["interfaces"][interface.name]["l2"] = {}

                # Get Native VLAN
                if interface.native_vlan or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["native_vlan"] = interface.native_vlan

                # Get Trunk Encapsulation
                if interface.trunk_encapsulation or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["trunk_encapsulation"] = interface.trunk_encapsulation

                # Get Switchport Mode
                if interface.switchport_mode or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["mode"] = interface.switchport_mode

                # Get Trunk Allowed VLANs
                if interface.trunk_allowed_vlans or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["allowed_vlans"] = interface.trunk_allowed_vlans

                # Get Access VLAN
                if interface.access_vlan or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["access_vlan"] = interface.access_vlan

                # Get Switchport Nonegotiate
                if interface.switchport_nonegotiate or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["switchport_nonegotiate"] = interface.switchport_nonegotiate
                
                # Get Voice VLAN
                if interface.voice_vlan or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["voice_vlan"] = interface.voice_vlan

                # Get Storm Control
                if interface.storm_control or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["storm_control"] = interface.storm_control

                # Get Device Tracking Policy
                if interface.device_tracking_policy or not self.omit_empty:
                    self.data["interfaces"][interface.name]["l2"]["device_tracking"] = {"policy": interface.device_tracking_policy}

            if "tunnel" in flags:
                self.data["interfaces"][interface.name]["tunnel"] = interface.tunnel_properties

    def parse_common(self):
        # Get Hostname
        self.data["hostname"] = self.config.hostname
        # Get Domain name
        self.data["domain_name"] = self.config.domain_name
        # Get Name-Servers
        self.data["name_servers"] = self.config.name_servers
        self.data["cdp"] = self.config.cdp
        self.data["vlans"] = self.config.vlans
        self.data["vrfs"] = self.config.vrfs
        try:
            if self.config.ntp or not self.omit_empty:
                self.data["ntp"] = self.config.ntp
            if self.config.logging_servers or not self.omit_empty:
                self.data["logging"] = self.config.logging
            if self.config.tacacs_servers or not self.omit_empty:
                self.data["tacacs_servers"] = self.config.tacacs_servers
            if self.config.radius_servers or not self.omit_empty:
                self.data["radius_servers"] = self.config.radius_servers
            if self.config.tacacs_groups or not self.omit_empty:
                self.data["tacacs_groups"] = self.config.tacacs_groups
            if self.config.radius_groups or not self.omit_empty:
                self.data["radius_groups"] = self.config.radius_groups
        except Exception as e:
            self.logger.error("Encountered Exception: {}".format(repr(e)))

    def get_interface_list(self, flags_filter=None):
        interfaces = []
        if not flags_filter:
            interfaces = list(self.data["interfaces"].keys())
        else:
            if isinstance(flags_filter, list):
                for interface, params in self.data["interfaces"].items():
                    for flag in flags_filter:
                        if flag in params["flags"]:
                            if interface not in interfaces:
                                interfaces.append(interface)
        return interfaces

    def get_ordered_interfaces(self):
        """
        Return interfaces as OrderedDict

        Returns:
            (:obj:`OrderedDict`): Interface section as OrderedDict

        """
        interfaces_crange = CiscoRange(list(self.data["interfaces"].keys()))
        ordered_interfaces = OrderedDict(sorted(self.data["interfaces"].items(), key=lambda x: interface_sort(crange=interfaces_crange, name=x[0])))

        return ordered_interfaces

    def to_json(self, indent=2):
        """
        Return JSON formatted structure describing configuration

        Args:
            indent (int): Set JSON indent, defaults to 2

        Returns:
            str: JSON string

        """
        # Convert Interfaces to OrderedDict
        data = dict(self.data)
        data["interfaces"] = self.get_ordered_interfaces()
        return json.dumps(obj=data, indent=indent)

    def to_yaml(self):
        """
        Return YAML formatted structure describing configuration

        Returns:
            str: YAML string
        """
        try:
            import yaml

            class CustomDumper(yaml.Dumper):

                def increase_indent(self, flow=False, indentless=False):
                    return super(CustomDumper, self).increase_indent(flow=flow, indentless=False)

        except ImportError:
            self.logger.error("Missing Package PyYAML. Please install it by running 'pip3 install pyyaml'")
            return ""

        # Convert Interfaces to OrderedDict
        data = dict(self.data)
        data["interfaces"] = self.get_ordered_interfaces()
        if has_old_pyyaml():
            data["interfaces"] = UnsortableOrderedDict(data["interfaces"])
            CustomDumper.add_representer(UnsortableOrderedDict, yaml.representer.SafeRepresenter.represent_dict)
            return yaml.dump(data=data, sort_keys=False, Dumper=CustomDumper)
        else:
            CustomDumper.add_representer(OrderedDict, yaml.representer.SafeRepresenter.represent_dict)
            return yaml.dump(data=data, sort_keys=False, Dumper=CustomDumper)

    @staticmethod
    def jprint(data):
        print(json.dumps(data, indent=2))

    



import pathlib
import os
import re
import json
import timeit
from ccutils.utils.common_utils import get_logger
from ccutils.ccparser import BaseConfigLine
from ccutils.ccparser import BaseInterfaceLine
from ccutils.utils import CiscoRange
import functools

re._MAXCACHE = 1024


class BaseConfigParser(object):
    """
    Base Configuration Parser object used for loading configuration files.
    """

    PATTERN_TYPE = type(re.compile(pattern=""))
    INTERFACE_LINE_CLASS = BaseInterfaceLine

    # Patterns
    _vlan_configuration_regex = re.compile(pattern=r"^vlan configuration (?P<vlan_range>[\d\-,]+)", flags=re.MULTILINE)
    _device_tracking_attach_policy_regex = re.compile(pattern=r"^ device-tracking attach-policy (?P<policy>\S+)")

    def __init__(self, config=None, verbosity=4, name="BaseConfigParser", **kwargs):
        """
        Base class for parsing Cisco-like configs

        Args:
            config (:obj:`pathlib.Path` or `str` or `list`): Config file in a form of `pathlib.Path`, or `string`
                containing the entire config or list of lines of the config file
            verbosity (:obj:`int`, optional): Determines the verbosity of logging output, defaults to 4: Info

        Attributes:
            lines (list): Contains list of all config lines stored as objects (see :class:`ccutils.ccparser.BaseConfigLine`)
            config_lines_str (list): Contains list of all config lines stored as strings

        Examples:

            Possible config inputs::

                # Using pathlib
                config_file = pathlib.Path("/path/to/config_file.txt")
                config = BaseConfigParser(config=config_file)

                # Using string
                config_string = '''
                hostname RouterA
                !
                interface Ethernet0/0
                 description Test Interface
                 ip address 10.0.0.1 255.255.255.0
                !
                end
                '''
                config = BaseConfigParser(config=config_string)

                # Using list
                config_list = [
                "hostname RouterA",
                    "!",
                    "interface Ethernet0/0",
                    " description Test Interface",
                    " ip address 10.0.0.1 255.255.255.0",
                    "!",
                    "end"
                ]
                config = BaseConfigParser(config=config_list)

        """
        self.verbosity = verbosity
        self.logger = get_logger(name=name, verbosity=verbosity)
        self.config = config
        self.path = self._check_path(kwargs.get("filepath", None)) if kwargs.get("filepath", None) else None

        self.minimal_results = True
        #: This is a URI.
        self.lines = []
        self.config_lines_str = []
        self.parse()

    @property
    def config_lines_obj(self):
        """
        Kept for backwards compatibility, will be removed in future versions.

        Returns:
            list: BaseConfigParser.lines

        """
        self.logger.warning("DEPRECATED: You are using deprecated property .config_lines_obj use .lines instead.")
        return self.lines

    def parse(self):
        """
        Entry function which triggers the parsing process. Called automatically when instantiating the object.

        :return: ``None``
        """
        if self.config:
            config_lines = []
            # Determine Config Type
            if isinstance(self.config, list):
                self.logger.debug(msg="Treating config as list of config lines.")
                config_lines = [x for x in self.config if (isinstance(x, str) and x != "")]
            elif isinstance(self.config, str):
                path = None
                if os.path.exists(self.config):
                    path = pathlib.Path(self.config)
                if path and path.exists():
                    self.logger.debug(msg="Treating config as filepath.")
                    path = self._check_path(filepath=path)
                    if path:
                        config_lines = [x for x in path.read_text().split("\n") if x != ""]
                else:
                    self.logger.debug(msg="Treating config as multi-line string.")
                    config_lines = [x for x in self.config.split("\n") if x != ""]
            elif isinstance(self.config, pathlib.Path):
                self.logger.debug(msg="Treating config as pathlib Path.")
                path = self._check_path(filepath=self.config)
                if path:
                    config_lines = [x for x in path.read_text().split("\n") if x != ""]
            else:
                self.logger.error("Invalid value passed as config argument.")
                raise ValueError("Invalid value passed as config argument.")
            self.config_lines_str = config_lines
            self.fix_indents()
            self._create_cfg_line_objects()
        else:
            self._get_clean_config()
            self.fix_indents()
            self._create_cfg_line_objects()

    def _check_path(self, filepath):
        path = None
        if not isinstance(filepath, pathlib.Path):
            path = pathlib.Path(filepath)
        else:
            path = filepath
        if not path.exists():
            self.logger.error(msg="Path '{}' does not exist.".format(filepath))
            return None
        if not path.is_file():
            self.logger.error(msg="Path '{}' is not a file.".format(filepath))
        if not path.is_absolute():
            path = path.resolve()
            self.logger.debug("Path '{}' is existing file.".format(filepath))
            return path
        else:
            self.logger.debug("Path '{}' is existing file.".format(filepath))
            return path

    def _get_indent(self, line):
        indent_size = len(line) - len(line.lstrip(" "))
        return indent_size

    def _get_clean_config(self, first_line_regex=r"^version \d+\.\d+", last_line_regex=r"^end"):
        first_regex = re.compile(pattern=first_line_regex, flags=re.MULTILINE)
        last_regex = re.compile(pattern=last_line_regex, flags=re.MULTILINE)
        all_lines = self.path.read_text().split("\n")
        first = None
        last = None
        for i in range(len(all_lines)):
            if not first:
                if re.match(pattern=first_regex, string=all_lines[i]):
                    first = i
                    self.logger.debug(msg="Found first config line: '{}'".format(all_lines[first]))
            if not last:
                if re.match(pattern=last_regex, string=all_lines[i]):
                    last = i
                    self.logger.debug(msg="Found last config line: '{}'".format(all_lines[last]))
                    break
        if not first or not last:
            self.config_lines_str = []
            self.logger.error(msg="No valid config found!")
        else:
            self.config_lines_str = all_lines[first:last + 1]
            self.logger.info(msg="Loading {} config lines.".format(len(self.config_lines_str)))
        # Fix indent

    def fix_indents(self):
        """
        Function for fixing the indentation level of config lines.

        :return:
        """
        indent_map = list(map(self._get_indent, self.config_lines_str))
        fixed_indent_map = []
        for i in range(len(indent_map)):
            if i == 0:
                fixed_indent_map.append(0)
                continue
            if indent_map[i] == 0:
                fixed_indent_map.append(0)
                continue
            if indent_map[i] == indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1])
            elif indent_map[i] > indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1]+1)
            elif indent_map[i] < indent_map[i-1]:
                fixed_indent_map.append(fixed_indent_map[-1]-1)
        for i, val in enumerate(fixed_indent_map):
            self.config_lines_str[i] = " "*val + self.config_lines_str[i].strip()
            #print(val, "'{}'".format(self.config_lines_str[i]))

    def _create_cfg_line_objects(self):
        """
        Function for generating ``self.lines``.

        """
        start = timeit.default_timer()
        for number, text in enumerate(self.config_lines_str):
            if re.match(pattern=r"^interface\s\S+", string=text, flags=re.MULTILINE):
                self.lines.append(self.INTERFACE_LINE_CLASS(number=number, text=text, config=self, verbosity=self.verbosity).return_obj())
            else:
                self.lines.append(BaseConfigLine(number=number, text=text, config=self, verbosity=self.verbosity).return_obj())
        for line in self.lines:
            line.type = line.get_type
        self.logger.debug(msg="Created {} ConfigLine objects in {} ms.".format(len(self.lines), (timeit.default_timer()-start)*1000))

    def _compile_regex(self, regex, flags=re.MULTILINE):
        """
        Helper function for compiling `re` patterns from string.

        :param str regex: Regex string
        :param flags: Flags for regex pattern, default is re.MULTILINE
        :return:
        """
        pattern = None
        try:
            pattern = re.compile(pattern=regex, flags=flags)
        except Exception as e:
            self.logger.error(msg="Error while compiling regex '{}'. Exception: {}".format(regex, repr(e)))
        return pattern

    def find_objects(self, regex, flags=re.MULTILINE):
        """
        Function for filtering Config Lines Objects based on given regex.

        Args:
            regex (:obj:`re.Pattern` or `str`): Regex based on which the search is done
            flags (:obj:`int`, optional): Set custom flags for regex, defaults to ``re.MULTILINE``

        Examples:

            Example::

                # Initialize the object
                config = BaseConfigParser(config="/path/to/config_file.cfg")

                # Define regex for matching config lines
                interface_regex = r"^ interface"

                # Apply the filter
                interface_lines = config.find_objects(regex=interface_regex)

                # Returns subset of ``self.lines`` which match specified regex

        """
        pattern = None
        if not isinstance(regex, self.PATTERN_TYPE):
            pattern = self._compile_regex(regex=regex, flags=flags)
        else:
            pattern = regex
        results = []
        for line in self.lines:
            if re.search(pattern=pattern, string=line.text):
                results.append(line)
        self.logger.debug(msg="Matched {} lines for query '{}'".format(len(results), regex))
        return results

    def get_section_by_parents(self, parents):
        if not isinstance(parents, list):
            parents = list(parents)
        section = list(self.lines)
        for parent in parents:
            section = [x.get_children() for x in section if bool(x.is_parent and x.re_match(parent))]
            if len(section) == 1:
                section = section[0]
            elif len(section) > 1:
                self.logger.error("Multiple lines matched parent statement. Cannot determine config section.")
                return []
            else:
                self.logger.error("No lines matched parent statement. Cannot determine config section.")
                return []
        return section

    def match_to_dict(self, line, patterns):
        """

        Args:
            line: Instance of `BaseConfigLine` object
            patterns: List of compiled `re` patterns
            minimal_result: Bool, if True, omits keys with value `None`

        Returns:
            dict: Dictionary containing named groups across all provided patterns

        """
        entry = {}

        for pattern in patterns:
            match_result = line.re_search(regex=pattern, group="ALL")
            if match_result is not None:
                entry.update(match_result)
            else:
                if self.minimal_results:
                    continue
                else:
                    entry.update({k: None for k in pattern.groupindex.keys()})
        return entry

    def property_autoparse(self, candidate_pattern, patterns):
        """
        Function for searching multiple patterns across all occurrences of lines that matched candidate_pattern
        Args:
            candidate_pattern:
            patterns:

        Returns:

        """
        properties = None
        candidates = self.find_objects(regex=candidate_pattern)
        if len(candidates):
            properties = []
        else:
            return properties
        for candidate in candidates:
            properties.append(self.match_to_dict(line=candidate, patterns=patterns))
        return properties

    def section_property_autoparse(self, parent, patterns, return_with_line=False):
        entries = None
        if isinstance(parent, BaseConfigLine):
            candidates = [parent]
        elif isinstance(parent, (str, self.PATTERN_TYPE)):
            candidates = self.find_objects(regex=parent)
        if len(candidates):
            entries = []
        else:
            return entries
        for candidate in candidates:
            entry = {}
            if isinstance(parent, (str, self.PATTERN_TYPE)):
                entry.update(self.match_to_dict(line=candidate, patterns=[parent]))
            for pattern in patterns:
                updates = candidate.re_search_children(regex=pattern, group="ALL")
                if len(updates) == 1:
                    entry.update(updates[0])
                elif len(updates) == 0:
                    if self.minimal_results:
                        continue
                    else:
                        entry.update({k: None for k in pattern.groupindex.keys()})
                else:
                    self.logger.warning("Multiple possible updates found for Pattern: '{}' on Candidate: '{}'".format(pattern, candidate))
            if return_with_line:
                entries.append((candidate, entry))
            else:
                entries.append(entry)
        return entries

    @property
    @functools.lru_cache()
    def hostname(self):
        hostname = None
        regex = r"^hostname\s(\S+)"
        candidates = self.find_objects(regex=regex)
        if len(candidates):
            hostname = candidates[0].re_search(regex=regex, group=1)
        return hostname

    @property
    @functools.lru_cache()
    def cdp(self):
        if len(self.find_objects(regex="^no cdp run")):
            return False
        else:
            return True

    @property
    def domain_name(self):
        domain_name = None
        domain_name_regex = re.compile(pattern=r"^ip domain.name (?P<domain_name>\S+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=domain_name_regex)
        if len(candidates):
            domain_name = candidates[0].re_search(regex=domain_name_regex, group="domain_name")
        return domain_name

    @property
    @functools.lru_cache()
    def name_servers(self):
        name_servers = []
        name_servers_regex = re.compile(pattern=r"^ip name.server (?P<name_servers>(?:\d{1,3}\.){3}\d{1,3}(?: (?:\d{1,3}\.){3}\d{1,3})*)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=name_servers_regex)
        for candidate in candidates:
            name_servers.extend(re.findall(pattern=r"(?:\d{1,3}\.){3}\d{1,3}", string=candidate.text))
        return name_servers

    @property
    def vlans(self):
        vlans = {}
        # Basic VLAN Definition
        vlan_id_regex = re.compile(pattern=r"^vlan (?P<vlan_id>\d+)", flags=re.MULTILINE)
        vlan_name_regex = re.compile(pattern=r"^ name (?P<vlan_name>\S+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=vlan_id_regex)
        for candidate in candidates:
            vlan_id = candidate.re_search(regex=vlan_id_regex, group="vlan_id")
            vlan_name = None
            vlan_name_candidate = candidate.re_search_children(regex=vlan_name_regex, group="vlan_name")
            if len(vlan_name_candidate):
                vlan_name = vlan_name_candidate[0]
            vlans[vlan_id] = {"name": vlan_name}
        # VLAN Configuration
        candidates = self.find_objects(regex=self._vlan_configuration_regex)
        for candidate in candidates:
            vlan_range = CiscoRange(candidate.re_search(regex=self._vlan_configuration_regex, group="vlan_range"))
            policy = candidate.re_search_children(regex=self._device_tracking_attach_policy_regex, group="policy")
            if len(policy):
                for vlan_id in vlan_range:
                    vlans[vlan_id]["device_tracking_policy"] = policy[0]
        return vlans

    @property
    def vlan_groups(self):
        vlan_group_regex = re.compile(pattern=r"^vlan group (?P<group>\S+) vlan-list (?P<vlan_id>\d+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=vlan_group_regex)
        return [x.re_search(regex=vlan_group_regex, group="ALL") for x in candidates]

    @property
    def vrfs(self):
        vrfs = None
        vrf_name_regex = re.compile(pattern=r"^(?:ip )?vrf(?: definition)? (?P<vrf_name>\S+)", flags=re.MULTILINE)
        rd_regex = re.compile(pattern=r"^ rd (?P<rd>\S+)", flags=re.MULTILINE)
        description_regex = re.compile(pattern=r"^ description (?P<description>.*)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=vrf_name_regex)
        if len(candidates):
            vrfs = {}
        for candidate in candidates:
            vrf_name = candidate.re_search(regex=vrf_name_regex, group="vrf_name")
            if vrf_name:
                vrfs[vrf_name] = {}
            else:
                continue
            rd_candidates = candidate.re_search_children(regex=rd_regex, group="rd")
            if len(rd_candidates):
                vrfs[vrf_name]["rd"] = rd_candidates[0]
            else:
                vrfs[vrf_name]["rd"] = None
            description_candidates = candidate.re_search_children(regex=description_regex, group="description")
            if len(description_candidates):
                vrfs[vrf_name]["description"] = description_candidates[0]
            else:
                vrfs[vrf_name]["description"] = None
        return vrfs

    @property
    def static_routes(self):
        # TODO: Add parsing for static routes
        pass

    @property
    def interface_lines(self):
        return (x for x in self.lines if x.is_interface)




import yaml
from yaml.representer import Representer
from yaml.dumper import Dumper
from yaml.emitter import Emitter
from yaml.serializer import Serializer
from yaml.resolver import Resolver
from collections import OrderedDict


def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)

class CustomAnsibleRepresenter(Representer):

    def represent_none(self, data):
        return self.represent_scalar(u'tag:yaml.org,2002:null', u'')


class CustomAnsibleDumper(Emitter, Serializer, CustomAnsibleRepresenter, Resolver):
    def __init__(self, stream,
                 default_style=None, default_flow_style=None,
                 canonical=None, indent=None, width=None,
                 allow_unicode=None, line_break=None,
                 encoding=None, explicit_start=None, explicit_end=None, sort_keys=False,
                 version=None, tags=None):
        Emitter.__init__(self, stream, canonical=canonical,
                         indent=indent, width=width,
                         allow_unicode=allow_unicode, line_break=line_break)
        Serializer.__init__(self, encoding=encoding,
                            explicit_start=explicit_start, explicit_end=explicit_end,
                            version=version, tags=tags)
        CustomAnsibleRepresenter.__init__(self, default_style=default_style,
                                          default_flow_style=default_flow_style)
        Resolver.__init__(self)

        CustomAnsibleRepresenter.add_representer(type(None), CustomAnsibleRepresenter.represent_none)
        CustomAnsibleRepresenter.add_representer(OrderedDict, represent_ordereddict)


    def increase_indent(self, flow=False, indentless=False):
        return super(CustomAnsibleDumper, self).increase_indent(flow=flow, indentless=False)


from ccutils.utils.common_utils import get_logger
from ccutils.utils.definitions import SCHEMA_DIR, SCHEMA_BASE_URI
import pathlib
import json
import jsonschema
import functools


class JsonValidator(object):

    def __init__(self, schema_dir=None, base_uri=None, verbosity=3):
        self.logger = get_logger(name="JsonValidator", verbosity=verbosity)
        self.schema_dir = pathlib.Path(schema_dir) if schema_dir is not None else SCHEMA_DIR
        self.logger.debug("Loading schemas from: {}".format(self.schema_dir))
        self.base_uri = base_uri if base_uri is not None else SCHEMA_BASE_URI
        self.logger.debug("Schema Base URI: {}".format(self.base_uri))
        self.schema_store = self.load_schema_store()
        self.logger.debug("Schema Store: {}".format(self.schema_store))
        self.test_schema_store()

    def load_schema_store(self):
        schema_store = {}

        for schema_path in [x for x in self.schema_dir.iterdir() if x.suffix == ".json"]:
            with schema_path.open() as schema_fp:
                try:
                    schema = json.load(fp=schema_fp)
                    if "$id" in schema.keys():
                        schema_store[schema["$id"]] = {"schema": schema}
                except Exception as e:
                    # TODO: Fix Exceptions
                    self.logger.error("Could not load valid JSON from: '{}' Exception: {}".format(schema_path, repr(e)))
        return schema_store

    @functools.lru_cache()
    def test_schema(self, schema_id):
        try:
            schema = self.schema_store["{}/{}".format(self.base_uri, schema_id)]["schema"]
            resolver = jsonschema.RefResolver(base_uri="file://{}/{}.json".format(self.schema_dir, schema_id), referrer=schema, store={k: v["schema"] for k, v in self.schema_store.items()})
            self.schema_store["{}/{}".format(self.base_uri, schema_id)]["resolver"] = resolver
            validator = jsonschema.Draft7Validator(schema=schema, resolver=resolver)
            self.schema_store["{}/{}".format(self.base_uri, schema_id)]["validator"] = validator
            validator.check_schema(schema=validator.schema)
            self.logger.debug("Schema '{}' is valid.".format(schema_id))
            return True
        except KeyError as e:
            self.logger.error("Schema '{}' not found in schema store. Exception: {}".format(schema_id, repr(e)))
            return False
        except jsonschema.exceptions.SchemaError as e:
            self.logger.error("Schema Error on '{}'. Exception: {}".format(schema_id, repr(e)))
        except Exception as e:
            self.logger.critical("Unhandled Exception occurred when validating schema: '{}' Exception: {}".format(schema_id, repr(e)))
            return False

    def validate(self, data, schema_id):
        result = False
        validator = self.schema_store["{}/{}".format(self.base_uri, schema_id)]["validator"]
        try:
            self.logger.debug("Validating data against schema: {}. Data: '{}'".format(schema_id, data))
            validator.validate(data)
            self.logger.debug("Data is valid acording to schema: {}. Data: '{}'".format(schema_id, data))
            result = True
        except jsonschema.exceptions.ValidationError as e:
            self.logger.warning("ValidationError: Data is not valid against the schema: {}. Data: '{}' Exception: {}".format(schema_id, data, repr(e)))
            result = False
        except Exception as e:
            self.logger.critical("Unhandled Exception occurred when validating data with schema: '{}' Exception: {}".format(schema_id, repr(e)))
            result = False
        finally:
            return result

    def test_schema_store(self):
        for schema_id in [x.split("/")[-1] for x in self.schema_store.keys()]:
            self.logger.debug("Validating schema: '{}'".format(schema_id))
            result = self.test_schema(schema_id=schema_id)
            if result:
                pass
            else:
                self.logger.error("Schema '{}' is NOT valid.".format(schema_id))




from ccutils.ccparser.BaseConfigLine import BaseConfigLine
from ccutils.ccparser.BaseInterfaceLine import BaseInterfaceLine
from ccutils.ccparser.BaseConfigParser import BaseConfigParser
from ccutils.ccparser.ConfigToJson import ConfigToJson
from ccutils.ccparser.ConfigMigration import ConfigMigration
from ccutils.ccparser.CiscoIosInterfaceLine import CiscoIosInterfaceLine
from ccutils.ccparser.CiscoIosParser import CiscoIosParser


def ConfigParser(config, device_type, verbosity=4):
    """
    Factory function for getting Parser object
    Args:
        config:
        device_type:
        verbosity:

    Returns:
        obj: Instance of proper Parsing class based on device_type

    """
    parser_class = BaseConfigParser
    if device_type == "ios":
        parser_class = CiscoIosParser

    return parser_class(config=config, verbosity=verbosity)


import pathlib
from ccutils.utils.common_utils import get_logger, load_excel_sheet, jprint, split_interface_name
from ccutils.ccparser import BaseConfigParser, ConfigToJson
import json


class ConfigMigration:
    def __init__(self, hostname, excel_path, excel_sheet, old_config_folder, verbosity=1):
        self.hostname = hostname
        self.logger = get_logger(name="ConfigMigration - {}".format(self.hostname), verbosity=verbosity)
        self.old_config_folder = old_config_folder if isinstance(old_config_folder, pathlib.Path) else pathlib.Path(old_config_folder)
        self.excel_data = load_excel_sheet(file=excel_path, sheet_name=excel_sheet)
        if not self.excel_data:
            self.logger.error(msg="Could not load Excel Data")
        self.old_hostnames = self.get_old_hostnames()
        self.old_configs = None
        self.old_ctj = None
        self.interface_map = self.get_interface_mapping()

        self.get_old_configs()
        self.get_old_ctj()


    def get_old_hostnames(self, column="Old Host"):
        old_hosts = set()
        if self.excel_data:
            try:
                for old_host in filter(lambda x: bool(x), self.excel_data[column]):
                    old_hosts.add(old_host)
            except Exception as e:
                self.logger.error(msg="Could not retrieve Old Hostnames from Excel Table. Reason: Unkown. Exception: {}".format(repr(e)))
        else:
            self.logger.error(msg="Could not retrieve Old Hostnames from Excel Table. Reason: File not loaded.")
        return list(old_hosts)
    
    def get_interface_mapping(self):
        interface_map = {}
        for new_interface in filter(lambda x: bool(x), self.excel_data["New Interface"]):
            i = self.excel_data["New Interface"].index(new_interface)
            if self.excel_data["Old Interface"][i] and self.excel_data["Old Host"][i]:
                #print("{} maps to {} on {}".format(new_interface, self.excel_data["Old Interface"][i], self.excel_data["Old Host"][i]))
                interface_map[new_interface] = [self.excel_data["Old Host"][i], self.excel_data["Old Interface"][i]]
            else:
                interface_map[new_interface] = None
        return interface_map


    def get_old_configs(self):
        old_configs = {}
        for old_hostname in self.old_hostnames:
            file_candidates = list(filter(lambda x: old_hostname in x.stem, self.old_config_folder.iterdir()))
            if len(file_candidates) == 1:
                self.logger.info(msg="Found config file for hostname: {}".format(old_hostname))
                old_configs[old_hostname] = BaseConfigParser(filepath=file_candidates[0])
            elif len(file_candidates) == 0:
                self.logger.error(msg="Could not find config file for hostname: {}".format(old_hostname))
            else:
                self.logger.error(msg="Found multiple configs for hostname: {}".format(old_hostname))
        self.old_configs = old_configs

    def get_old_ctj(self):
        old_ctj = {}
        for old_hostname, config in self.old_configs.items():
            old_ctj[old_hostname] = ConfigToJson(config=config)
        self.old_ctj = old_ctj


    def get_new_interface(self, old_host, old_interface):
        new_interface = None
        if old_interface in self.old_ctj[old_host].data["interfaces"].keys():
            self.logger.debug(msg="Requested interface {} exists in config of host {}".format(old_interface, old_host))
        candidates = [x for x in self.interface_map.keys() if (self.interface_map[x] and (self.interface_map[x][0] == old_host and self.interface_map[x][1] == old_interface))]
        if len(candidates) == 1:
            new_interface = candidates[0]
            self.logger.debug(msg="New interface for {} / {} is {}".format(old_host, old_interface, new_interface))
        else:
            self.logger.error(msg="No mapping for interface found. ({} / {})".format(old_host, old_interface))
        return new_interface


    def get_context_for_new_interface(self, new_interface):
        old_interface = None
        old_host = None
        new_interface_context = None
        if new_interface not in self.interface_map.keys():
            self.logger.error(msg="No such New Interface Exists: {}".format(new_interface))
            
        elif not self.interface_map[new_interface]:
            self.logger.warning(msg="No Old Interface exists for New Interface: {}".format(new_interface))
            
        else:
            old_host = self.interface_map[new_interface][0]
            old_interface = self.interface_map[new_interface][1]
            old_interface_context = self.old_ctj[old_host].data["interfaces"][old_interface]
            # Replace Etherchannel group Number
            if old_interface_context["channel_group"]:
                old_channel_group_number = old_interface_context["channel_group"]["channel_group_number"]
                old_portchannel = "Port-channel{}".format(old_channel_group_number)
                new_portchannel = self.get_new_interface(old_host=old_host, old_interface=old_portchannel)
                if new_portchannel:
                    new_channel_group_number = split_interface_name(new_portchannel)[1]
                    new_interface_context = old_interface_context
                    new_interface_context["channel_group"]["channel_group_number"] = new_channel_group_number
                else:
                    self.logger.error(msg="Could not determine new Port-channel number. No replacement for interface {} / {} exists.".format(old_host, old_portchannel))
                    
            else:
                new_interface_context = old_interface_context
        return new_interface_context


    def check_standby(self):
        groups = {}
        for old_hostname, old_ctj in self.old_ctj.items():
            for interface, params in old_ctj.data["interfaces"].items():
                if "standby" in params["flags"]:
                    for group in params["l3"]["standby"]["groups"].keys():
                        if group not in groups.keys():
                            groups[group] = []
                        groups[group] += old_hostname, interface, params["l3"]["standby"]["groups"][group]["priority"]
        
        for group, value in groups.items():
            line = [group]
            for hostname in self.old_hostnames:
                if hostname in value:
                    line += value[value.index(hostname)], value[value.index(hostname)+1]
                else:
                    line += "", ""

            print(";".join(line))
    
    def user_selection(self, prompt, options):
        result = int(input(prompt))
        return options[result]

    def merge_vlans(self):
        vlans = {}
        for hostname, old_ctj in self.old_ctj.items():
            for vlan, params in old_ctj.data["vlans"].items():
                if vlan not in vlans.keys():
                    vlans[vlan] = params
                    self.logger.debug(msg="Added new VLAN: {} '{}'".format(vlan, params["name"]))
                else:
                    if vlans[vlan]["name"] == params["name"]:
                        pass
                    elif vlans[vlan]["name"] != params["name"]:
                        if not vlans[vlan]["name"]:
                            vlans[vlan]["name"] = params["name"]
                            self.logger.debug(msg="Renamed VLAN {} to '{}'".format(vlan, params["name"]))
                        elif not params["name"]:
                            pass
                        else: 
                            options = {1: vlans[vlan]["name"], 2: params["name"]}
                            result = self.user_selection(prompt="Found two diffenet names for same VLAN {}: {}:\n".format(vlan, options), options=options)
                            vlans[vlan]["name"] = result
        
        return vlans
    
    def merge_vrfs(self):
        vrfs = {}
        for hostname, old_ctj in self.old_ctj.items():
            self.logger.info(msg="Proccessing VRFs on {}".format(hostname))
            for vrf, params in old_ctj.data["vrfs"].items():
                if vrf not in vrfs.keys():
                    vrfs[vrf] = params
                    self.logger.debug(msg="Added new VRF: {}".format(vrf))
                else:
                    self.logger.info(msg="VRF {} exists.".format(vrf))
                    if vrfs[vrf]["rd"] == params["rd"]:
                        self.logger.info(msg="VRF has same RD.")
                        pass
                    elif vrfs[vrf]["rd"] != params["rd"]:
                        self.logger.info(msg="VRF does NOT have same RD.")
                        if not vrfs[vrf]["rd"]:
                            vrfs[vrf]["rd"] = params["rd"]
                        elif not params["rd"]:
                            pass
                        else:
                            options = {1: vrfs[vrf]["rd"], 2: params["rd"]}
                            result = self.user_selection(prompt="Found two diffenet RDs for same VRF {}: {}:\n".format(vrf, options), options=options)
                            vrfs[vrf]["rd"] = result
                    
                    if vrfs[vrf]["description"] == params["description"]:
                        pass
                    elif vrfs[vrf]["description"] != params["description"]:
                        if not vrfs[vrf]["description"]:
                            vrfs[vrf]["description"] = params["description"]
                        elif not params["description"]:
                            pass
                        else:
                            options = {1: vrfs[vrf]["description"], 2: params["description"]}
                            result = self.user_selection(prompt="Found two diffenet descriptions for same VRF {}: {}:\n".format(vrf, options), options=options)
                            vrfs[vrf]["description"] = result

        return vrfs




from setuptools import setup, find_packages

with open("requirements.txt", "r") as fs:
    reqs = [r for r in fs.read().splitlines() if (len(r) > 0 and not r.startswith("#"))]

setup(
    name='ccutils',
    version='0.2.18',
    packages=find_packages(exclude=["test", "examples"]),
    url='https://github.org/mihudec/ccutils',
    license='',
    author='Miroslav Hudec',
    author_email='mijujda@gmail.com',
    description='Cisco Configuration Utilities',
    install_requires=reqs,
    include_package_data=True
)



from ccutils.ccparser import BaseConfigParser
from ccutils.ccparser import CiscoIosInterfaceLine
from ccutils.utils import CiscoRange
from ccutils.utils.common_utils import remove_empty_values, value_to_bool
import functools
import re


class CiscoIosParser(BaseConfigParser):

    INTERFACE_LINE_CLASS = CiscoIosInterfaceLine


    # Regexes
    _interface_pattern = r"[A-z]{2,}(?:[A-z\-])?\d+(?:\/\d+)?(?:\:\d+)?(?:\.\d+)?"
    _ip_address_pattern = r"(?:\d{1,3}\.){3}\d{1,3}"
    _host_pattern = r"[A-z0-9\-\_\.]+"
    _acl_name = r"[A-z0-9\_\-\.]+"

    _vlan_configuration_regex = re.compile(pattern=r"^vlan configuration (?P<vlan_range>[\d\-,]+)", flags=re.MULTILINE)
    _device_tracking_attach_policy_regex = re.compile(pattern=r"^ device-tracking attach-policy (?P<policy>\S+)")

    _source_interface_regex = re.compile(pattern=r"source (?P<source_interface>{0})".format(_interface_pattern))
    _source_vrf_regex = re.compile(pattern=r"vrf (?P<vrf>\S+)")

    _name_server_base_regex = re.compile(pattern=r"^ip name.server (?P<name_servers>(?:\d{1,3}\.){3}\d{1,3}(?: (?:\d{1,3}\.){3}\d{1,3})*)", flags=re.MULTILINE)

    _ntp_server_base_regex = re.compile(pattern=r"^ntp server(?: vrf \S+)? (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern))
    _ntp_peer_base_regex = re.compile(pattern=r"^ntp peer(?: vrf \S+)? (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern))
    _ntp_access_group_regex = re.compile(pattern=r"^ntp access-group (?P<access_type>peer|serve|serve-only|query-only) (?P<acl>{0})".format(_acl_name))
    _ntp_authentication_keys_regex = re.compile(pattern=r"^ntp authentication-key (?P<key>\d+) (?P<hash_algorithm>\S+) (?P<hash>\S+)(?: (?P<encryption_type>\d+))?".format(_acl_name))
    _ntp_trusted_key_regex = re.compile(pattern=r"ntp trusted-key (?P<key>\d+)")
    _ntp_source_regex = re.compile(pattern="^ntp source (?P<source>{0})".format(_interface_pattern))

    _logging_source_interface_regex = re.compile(pattern=r"^logging source-interface (?P<source>{0})".format(_interface_pattern))
    _logging_server_base_regex = re.compile(pattern=r"^logging host (?P<server>{0}|{1})".format(_ip_address_pattern, _host_pattern))
    _logging_transport_regex = re.compile(pattern=r"transport (?P<protocol>udp|tcp) port (?P<port>\d+)")

    _tacacs_server_regex = re.compile(pattern=r"^tacacs server (?P<name>\S+)")
    _radius_server_regex = re.compile(pattern=r"^radius server (?P<name>\S+)")

    _radius_auth_port_regex = re.compile(pattern=r"auth-port (?P<auth_port>\d+)")
    _radius_acct_port_regex = re.compile(pattern=r"acct-port (?P<acct_port>\d+)")

    _aaa_radius_group_regex = re.compile(pattern=r"^aaa group server radius (?P<name>\S+)")
    _aaa_tacacs_group_regex = re.compile(pattern=r"^aaa group server tacacs\+ (?P<name>\S+)")

    _aaa_group_server_name_regex = re.compile(pattern=r"^ server name (?P<name>\S+)")
    _aaa_group_vrf_regex = re.compile(pattern=r"^ ip vrf forwarding (?P<vrf>\S+)")
    _aaa_group_source_interface_regex = re.compile(pattern=r"^ ip (?:tacacs|radius) source-interface (?P<source_interface>{0})".format(_interface_pattern))

    _aaa_server_address_regex = re.compile(pattern=r"^ address (?P<address_version>ipv4|ipv6) (?P<server>\S+)")
    _aaa_server_key_regex = re.compile(pattern=r"^ key(?: (?P<encryption_type>\d+))? (?P<hash>\S+)")
    _aaa_server_timeout_regex = re.compile(pattern=r"^ timeout (?P<timeout>\d+)")
    _aaa_server_retransmit_regex = re.compile(pattern=r"^ retransmit (?P<retransmit>\d+)")
    _aaa_server_single_connection = re.compile(pattern=r"^ (?P<single_connection>single-connection)")

    _aaa_methods_regex = re.compile(pattern=r"(?:group \S+)|local|radius|tacacs\+")
    _aaa_authentication_login_regex = re.compile(pattern=r"^aaa authentication login (?P<name>\S+)")
    _aaa_authorization_exec_regex = re.compile(pattern=r"^aaa authorization exec (?P<name>\S+)")

    _routing_ospf_process_regex = re.compile(pattern=r"^router ospf (?P<process_id>)( (?P<vrf>\S+))?")
    _routing_isis_process_regex = re.compile(pattern=r"^router isis (?P<process_id>\S+)")
    _routing_isis_network_id_regex = re.compile(pattern=r"^ net (?P<area_id>\d{2}\.\d{4})\.(?P<system_id>(?:\d{4}\.){2}\d{4})\.(?P<nsel>\d{2})")
    _routing_isis_authentication_mode_candidates_regex = re.compile(pattern=r"^ authentication mode (?P<auth_mode>\S+) (?P<level>level-.)")
    _routing_isis_authentication_keychain_candidates_regex = re.compile(pattern=r"^ authentication key-chain (?P<keychain>\S+) (?P<level>level-.)")

    _vrf_definition_regex = re.compile(pattern=r"^(?:ip )?vrf(?: definition)? (?P<name>\S+)")
    _vrf_rd_regex = re.compile(pattern=r"rd (?P<rd>\d+:\d+)")
    _address_family_regex = re.compile(pattern=r"address-family (?P<afi>\S+)(?:(?P<vrf>\S+))?")
    _description_regex = re.compile(pattern=r"description (?P<description>.*?)$")
    _vrf_afi_rt_regex = re.compile(pattern=r"route-target (?P<action>import|export) (?P<rt>\d+:\d+)")

    def __init__(self, config=None, verbosity=4, **kwargs):
        super(CiscoIosParser, self).__init__(config=config, verbosity=verbosity, name="CiscoIosParser", **kwargs)

    @property
    @functools.lru_cache()
    def hostname(self):
        hostname = None
        regex = r"^hostname\s(\S+)"
        candidates = self.find_objects(regex=regex)
        if len(candidates):
            hostname = candidates[0].re_search(regex=regex, group=1)
        return hostname

    @property
    @functools.lru_cache()
    def cdp(self):
        if len(self.find_objects(regex="^no cdp run")):
            return False
        else:
            return True

    @property
    def domain_name(self):
        domain_name = None
        domain_name_regex = re.compile(pattern=r"^ip domain.name (?P<domain_name>\S+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=domain_name_regex)
        if len(candidates):
            domain_name = candidates[0].re_search(regex=domain_name_regex, group="domain_name")
        return domain_name

    @property
    @functools.lru_cache()
    def name_servers(self):
        name_servers = None
        candidates = self.find_objects(regex=self._name_server_base_regex)
        if len(candidates):
            name_servers = []
        else:
            return name_servers
        for candidate in candidates:
            name_servers.extend(re.findall(pattern=r"(?:\d{1,3}\.){3}\d{1,3}", string=candidate.text))
        return name_servers

    @property
    @functools.lru_cache()
    def ntp_servers(self):
        """
        Property containing DNS servers related data

        Returns:
            list: List of name server IP addresses

            Example::

                [
                    "10.0.0.1",
                    "10.0.0.2"
                ]

            Returns ``None`` if absent


        """
        candidate_pattern = self._ntp_server_base_regex
        patterns = [
            self._ntp_server_base_regex,
            self._source_interface_regex,
            self._source_vrf_regex,
            re.compile("key (?P<key>\d+)"),
            re.compile("(?P<prefer>prefer)")
        ]
        ntp_servers = self.property_autoparse(candidate_pattern=candidate_pattern, patterns=patterns)
        if ntp_servers is not None:
            for server in ntp_servers:
                server = value_to_bool(entry=server, keys=["prefer"])
        return ntp_servers

    @property
    @functools.lru_cache()
    def ntp_peers(self):
        candidate_pattern = self._ntp_peer_base_regex
        patterns = [
            self._ntp_peer_base_regex,
            self._source_interface_regex,
            self._source_vrf_regex,
            re.compile("key (?P<key>\d+)")
        ]
        ntp_peers = self.property_autoparse(candidate_pattern=candidate_pattern, patterns=patterns)
        return ntp_peers

    @property
    @functools.lru_cache()
    def ntp_access_groups(self):
        ntp_access_groups = None
        candidate_pattern = self._ntp_access_group_regex
        candidates = self.find_objects(regex=candidate_pattern)
        if len(candidates):
            ntp_access_groups = {"peer": None, "serve": None, "serve-only": None, "query-only": None}
        else:
            return ntp_access_groups
        for candidate in candidates:
            match_result = candidate.re_search(regex=candidate_pattern, group="ALL")
            if match_result is not None:
                ntp_access_groups.update({match_result["access_type"]: match_result["acl"]})
        return ntp_access_groups

    @property
    @functools.lru_cache()
    def ntp_authentication_keys(self):
        candidate_pattern = self._ntp_authentication_keys_regex
        patterns = [
            self._ntp_authentication_keys_regex
        ]
        ntp_authentication_keys = self.property_autoparse(candidate_pattern=candidate_pattern, patterns=patterns)
        return ntp_authentication_keys

    @property
    @functools.lru_cache()
    def ntp_trusted_keys(self):
        candidate_pattern = self._ntp_trusted_key_regex
        patterns = [
            self._ntp_trusted_key_regex
        ]
        ntp_trusted_keys = self.property_autoparse(candidate_pattern=candidate_pattern, patterns=patterns)
        return ntp_trusted_keys

    @property
    @functools.lru_cache()
    def ntp_global_params(self):
        ntp_global_params = {
            "source": None,
            "authenticate": None
        }
        source_candidates = self.find_objects(regex=self._ntp_source_regex)
        if len(source_candidates) == 1:
            ntp_global_params["source"] = source_candidates[0].re_search(regex=self._ntp_source_regex, group="source")

        authenticate_candidates = self.find_objects(regex=r"^ntp authenticate$")
        if len(authenticate_candidates):
            ntp_global_params["authenticate"] = True
        return ntp_global_params

    @property
    @functools.lru_cache()
    def ntp(self):
        ntp = {}
        ntp.update(self.ntp_global_params)
        ntp["servers"] = self.ntp_servers
        ntp["peers"] = self.ntp_peers
        ntp["access_groups"] = self.ntp_access_groups
        ntp["authentication_keys"] = self.ntp_authentication_keys
        ntp["trusted_keys"] = self.ntp_trusted_keys
        is_empty = True
        for value in ntp.values():
            if value is not None:
                is_empty = False
        if is_empty:
            ntp = None
        return ntp

    @property
    @functools.lru_cache()
    def logging_servers(self):
        candidate_pattern = self._logging_server_base_regex
        patterns = [
            self._logging_server_base_regex,
            self._source_interface_regex,
            self._source_vrf_regex,
            self._logging_transport_regex
        ]
        logging_servers = self.property_autoparse(candidate_pattern=candidate_pattern, patterns=patterns)
        return logging_servers

    @property
    @functools.lru_cache()
    def logging_global_params(self):
        logging_global_params = {
            "sources": None
        }
        patterns = [
            self._logging_source_interface_regex,
            self._source_vrf_regex
        ]
        logging_global_params["sources"] = self.property_autoparse(candidate_pattern=self._logging_source_interface_regex, patterns=patterns)

        is_empty = True
        for value in logging_global_params.values():
            if value is not None:
                is_empty = False
        if is_empty:
            logging_global_params = None
        return logging_global_params

    @property
    @functools.lru_cache()
    def logging(self):
        logging = {}
        if self.logging_global_params is not None:
            logging.update(self.logging_global_params)
        logging["servers"] = self.logging_servers

        # Check for emptyness
        is_empty = True
        for value in logging.values():
            if value is not None:
                is_empty = False
                break
        if is_empty:
            logging = None
        return logging

    @property
    @functools.lru_cache()
    def tacacs_servers(self):
        """

        Returns:
            list: List of TACACS Servers

            Example::

                [
                    {
                      "name": "ISE-1",
                      "address_version": "ipv4",
                      "server": "10.0.0.1",
                      "encryption_type": "7",
                      "hash": "36A03A8A4C00E81F03D62D8B04BBBF4D",
                      "timeout": "10",
                      "single_connection": true
                    },
                    {
                      "name": "ISE-2",
                      "address_version": "ipv4",
                      "server": "10.0.1.1",
                      "encryption_type": "7",
                      "hash": "36A03A8A4C00E81F03D62D8B04BBBF4D",
                      "timeout": "10",
                      "single_connection": true
                    }
                ]

            Returns ``None`` if absent


        """
        patterns = [
            self._aaa_server_address_regex,
            self._aaa_server_key_regex,
            self._aaa_server_timeout_regex,
            self._aaa_server_retransmit_regex,
            self._aaa_server_single_connection
        ]
        tacacs_servers = self.section_property_autoparse(parent=self._tacacs_server_regex, patterns=patterns)
        if tacacs_servers is not None:
            for entry in tacacs_servers:
                entry = value_to_bool(entry=entry, keys=["single_connection"])
        return tacacs_servers

    @property
    @functools.lru_cache()
    def radius_servers(self):
        """

        Returns:
            list: List of RADIUS Servers

            Example::

                [
                    {
                      "name": "RADIUS-Primary",
                      "address_version": "ipv4",
                      "server": "10.0.0.1",
                      "encryption_type": null,
                      "hash": "Test123",
                      "timeout": "2",
                      "retransmit": "1",
                      "auth_port": "1812",
                      "acct_port": "1813"
                    },
                    {
                      "name": "RADIUS-Secondary",
                      "address_version": "ipv4",
                      "server": "10.0.1.1",
                      "encryption_type": null,
                      "hash": "Test123",
                      "timeout": "2",
                      "retransmit": "1",
                      "auth_port": "1812",
                      "acct_port": "1813"
                    }
                ]

            Returns ``None`` if absent

        """
        patterns = [
            self._aaa_server_address_regex,
            self._aaa_server_key_regex,
            self._aaa_server_timeout_regex,
            self._aaa_server_retransmit_regex,
            self._aaa_server_single_connection,
            self._radius_auth_port_regex,
            self._radius_acct_port_regex
        ]
        radius_servers = self.section_property_autoparse(parent=self._radius_server_regex, patterns=patterns)
        if radius_servers is not None:
            for entry in radius_servers:
                entry = value_to_bool(entry=entry, keys=["single_connection"])
        return radius_servers

    @property
    @functools.lru_cache()
    def tacacs_groups(self):
        """

        Returns:
            list: List of TACACS Server entries

            Example::

                [
                    {
                        "name": "ISE-TACACS",
                        "source_interface": "Loopback0",
                        "servers": [
                            {
                                "name": "ISE-1"
                            },
                            {
                                "name": "ISE-2"
                            }
                        ]
                    }
                ]

            Returns ``None`` if absent

        """
        patterns = [
            self._aaa_group_vrf_regex,
            self._aaa_group_source_interface_regex,
        ]
        tacacs_groups = self.section_property_autoparse(parent=self._aaa_tacacs_group_regex, patterns=patterns)
        if tacacs_groups is not None:
            candidates = self.find_objects(regex=self._aaa_tacacs_group_regex)
            for candidate in candidates:
                name = candidate.re_search(regex=self._aaa_tacacs_group_regex, group="name")
                servers = candidate.re_search_children(regex=self._aaa_group_server_name_regex, group="ALL")
                for tacacs_group in tacacs_groups:
                    if tacacs_group["name"] == name:
                        tacacs_group["servers"] = servers
        return tacacs_groups

    @property
    @functools.lru_cache()
    def radius_groups(self):
        """

        Returns:
            list: List of RADIUS Server Groups Entries

            Example::

                [
                    {
                        "name": "RADIUS-GROUP",
                        "source_interface": "Vlan100",
                        "servers": [
                            {
                                "name": "RADIUS-Primary"
                            }
                        ]
                    }
                ]

            Returns ``None`` if absent

        """
        patterns = [
            self._aaa_group_vrf_regex,
            self._aaa_group_source_interface_regex
        ]
        radius_groups = self.section_property_autoparse(parent=self._aaa_radius_group_regex, patterns=patterns)
        if radius_groups is not None:
            candidates = self.find_objects(regex=self._aaa_radius_group_regex)
            for candidate in candidates:
                name = candidate.re_search(regex=self._aaa_radius_group_regex, group="name")
                servers = candidate.re_search_children(regex=self._aaa_group_server_name_regex, group="ALL")
                for radius_group in radius_groups:
                    if radius_group["name"] == name:
                        radius_group["servers"] = servers
        return radius_groups

    @property
    @functools.lru_cache()
    def aaa_login_methods(self):
        aaa_login_methods = None
        candidates = self.find_objects(regex=self._aaa_authentication_login_regex)
        if len(candidates):
            aaa_login_methods = []
            for candidate in candidates:
                entry = self.match_to_dict(candidate, patterns=[self._aaa_authentication_login_regex])
                entry["methods"] = []
                methods = re.findall(pattern=self._aaa_methods_regex, string=candidate.text)
                for method in methods:
                    if method.startswith("group "):
                        entry["methods"].append(method[6:])
                    else:
                        entry["methods"].append(method)
                aaa_login_methods.append(entry)
        return aaa_login_methods

    @property
    @functools.lru_cache()
    def aaa_authorization_exec_methods(self):
        aaa_authorization_exec_methods = None
        candidates = self.find_objects(regex=self._aaa_authorization_exec_regex)
        if len(candidates):
            aaa_authorization_exec_methods = []
            for candidate in candidates:
                entry = self.match_to_dict(candidate, patterns=[self._aaa_authorization_exec_regex])
                entry["methods"] = []
                methods = re.findall(pattern=self._aaa_methods_regex, string=candidate.text)
                for method in methods:
                    if method.startswith("group "):
                        entry["methods"].append(method[6:])
                    else:
                        entry["methods"].append(method)
                aaa_authorization_exec_methods.append(entry)
        return aaa_authorization_exec_methods

    @property
    def vlans(self):
        vlans = {}
        # Basic VLAN Definition
        vlan_id_regex = re.compile(pattern=r"^vlan (?P<vlan_id>\d+)", flags=re.MULTILINE)
        vlan_name_regex = re.compile(pattern=r"^ name (?P<vlan_name>\S+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=vlan_id_regex)
        for candidate in candidates:
            vlan_id = candidate.re_search(regex=vlan_id_regex, group="vlan_id")
            vlan_name = None
            vlan_name_candidate = candidate.re_search_children(regex=vlan_name_regex, group="vlan_name")
            if len(vlan_name_candidate):
                vlan_name = vlan_name_candidate[0]
            vlans[vlan_id] = {"name": vlan_name}
        # VLAN Configuration
        candidates = self.find_objects(regex=self._vlan_configuration_regex)
        for candidate in candidates:
            vlan_range = CiscoRange(candidate.re_search(regex=self._vlan_configuration_regex, group="vlan_range"))
            policy = candidate.re_search_children(regex=self._device_tracking_attach_policy_regex, group="policy")
            if len(policy):
                for vlan_id in vlan_range:
                    # Fix for VLAN 1
                    try:
                        vlans[vlan_id]["device_tracking_policy"] = policy[0]
                    except KeyError as e:
                        if vlan_id == "1":
                            vlans["1"] = {"name": "default", "device_tracking_policy": policy[0]}
        return vlans

    @property
    def vlan_groups(self):
        vlan_group_regex = re.compile(pattern=r"^vlan group (?P<group>\S+) vlan-list (?P<vlan_id>\d+)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=vlan_group_regex)
        return [x.re_search(regex=vlan_group_regex, group="ALL") for x in candidates]

    @property
    def vrfs(self):
        vrfs = None
        vrf_name_regex = re.compile(pattern=r"^(?:ip )?vrf(?: definition)? (?P<vrf_name>\S+)", flags=re.MULTILINE)
        rd_regex = re.compile(pattern=r"^ rd (?P<rd>\S+)", flags=re.MULTILINE)
        description_regex = re.compile(pattern=r"^ description (?P<description>.*)", flags=re.MULTILINE)
        candidates = self.find_objects(regex=self._vrf_definition_regex)
        if len(candidates):
            vrfs = {}
        for candidate in candidates:
            vrf_name = candidate.re_search(regex=self._vrf_definition_regex, group="name")
            if vrf_name:
                vrfs[vrf_name] = {}
            else:
                continue
            rd_candidates = candidate.re_search_children(regex=self._vrf_rd_regex, group="ALL")
            if len(rd_candidates):
                vrfs[vrf_name].update(rd_candidates[0])
            else:
                vrfs[vrf_name]["rd"] = None
            description_candidates = candidate.re_search_children(regex=self._description_regex, group="ALL")
            if len(description_candidates):
                vrfs[vrf_name].update(description_candidates[0])
            else:
                vrfs[vrf_name]["description"] = None

            # Address Family Section
            address_families = []
            afi_lines = candidate.re_search_children(regex=self._address_family_regex)
            for line in afi_lines:
                entry = line.re_search(regex=self._address_family_regex, group="ALL")
                rt_candidates = line.re_search_children(regex=self._vrf_afi_rt_regex, group="ALL")
                if len(rt_candidates):
                    entry["import"] = [{"rt": x["rt"]} for x in rt_candidates if x["action"] == "import"]
                    entry["export"] = [{"rt": x["rt"]} for x in rt_candidates if x["action"] == "export"]
                entry = remove_empty_values(entry)
                address_families.append(entry)
            if len(address_families):
                vrfs[vrf_name]["address_families"] = address_families

        return vrfs

    @property
    def ospf(self):
        """

        Returns:

        """
        raise NotImplementedError()

    @property
    def isis(self):
        """

        Returns:

        """
        isis = []
        patterns = [
            self._routing_isis_process_regex,
            re.compile(pattern=r"^ is-type (?P<is_type>\S+)"),
            re.compile(pattern=r"^ metric-style (?P<metric_style>\S+)"),
            re.compile(pattern=r"^ fast-flood (?P<fast_flood>\S+)"),
            re.compile(pattern=r"^ max-lsp-lifetime (?P<max_lsp_lifetime>\S+)"),
        ]
        candidates = self.section_property_autoparse(parent=self._routing_isis_process_regex, patterns=patterns, return_with_line=True)
        if len(candidates):
            for parent, entry in candidates:
                net_candidates = parent.re_search_children(regex=self._routing_isis_network_id_regex, group="ALL")
                if len(net_candidates):
                    entry["net"] = net_candidates[0]
                authentication_mode_candidates = parent.re_search_children(regex=self._routing_isis_authentication_mode_candidates_regex, group="ALL")
                authentication_keychain_candidates = parent.re_search_children(regex=self._routing_isis_authentication_keychain_candidates_regex, group="ALL")
                if len(authentication_mode_candidates) or len(authentication_keychain_candidates):
                    entry["authentication"] = {}
                    entry["authentication"]["mode"] = authentication_mode_candidates
                    entry["authentication"]["keychain"] = authentication_keychain_candidates

                check_patterns = list(patterns)
                check_patterns.append(self._routing_isis_network_id_regex)
                check_patterns.append(self._routing_isis_authentication_mode_candidates_regex)
                check_patterns.append(self._routing_isis_authentication_keychain_candidates_regex)
                entry["unprocessed_lines"] = [x.text for x in self.section_unprocessed_lines(parent=parent, check_patterns=check_patterns)]


            isis.append(entry)
        else:
            isis = None
        return isis

    def section_unprocessed_lines(self, parent, check_patterns):
        unprocessed_lines = []
        for child in parent.get_children():
            processed = False
            for regex in check_patterns:
                match = child.re_search(regex=regex)
                if match:
                    processed = True
                    break
            if not processed:
                unprocessed_lines.append(child)
        return unprocessed_lines

    @property
    @functools.lru_cache()
    def all_ipv4_physical_addresses(self):
        """
        Get all physical IPv4 addresses of device config.

        Returns:
            list: List IPv4 Addresses

            Example::

                [
                    "192.168.100.2",
                    "192.168.101.2"
                ]

            Returns empty list (``[]``) if absent

        """
        addresses = []
        for interface_line in self.interface_lines:
            if "l3" in interface_line.flags:
                addresses.extend([x["address"] for x in interface_line.ipv4_addresses])
        return addresses

    @functools.lru_cache()
    def vrf_ipv4_physical_addresses(self, vrf="global"):
        """
        Get all physical IPv4 addresses in particular VRF. By default returns all IP addresses from global routing table

        Args:
            vrf: Name of the VRF (default is `"global"`)

        Returns:
            list: List IPv4 Addresses

            Example::

                [
                    "192.168.100.2",
                    "192.168.101.2"
                ]

            Returns empty list (``[]``) if absent

        """
        addresses = []
        # Filter L3 Interfaces
        interfaces = [x for x in self.interface_lines if "l3" in x.flags]
        # Filter VRF interfaces
        if vrf == "global":
            interfaces = [x for x in interfaces if x.vrf is None]
        else:
            interfaces = [x for x in interfaces if x.vrf == vrf]
        for interface in interfaces:
            addresses.extend([x["address"] for x in interface.ipv4_addresses])
        return addresses

    @property
    @functools.lru_cache()
    def all_ipv4_standby_addresses(self):
        """

        Returns:
            list: List IPv4 Addresses

            Example::

                [
                    "192.168.100.1",
                    "192.168.101.1"
                ]

            Returns empty list (``[]``) if absent

        """
        addresses = []
        # Filter L3 interfaces
        interfaces = [x for x in self.interface_lines if "l3" in x.flags]
        print(interfaces)
        # Filter Stadnby Interfaces
        interfaces = [x for x in interfaces if x.standby is not None]
        print(interfaces)
        for interface in interfaces:
            for group in interface.standby["groups"].values():
                print(group)
                addresses.extend([x["address"] for x in group["ipv4"]])
        return addresses

    @functools.lru_cache()
    def vrf_ipv4_standby_addresses(self, vrf="global"):
        """

        Args:
            vrf:  Name of the VRF (default is `"global"`)

        Returns:
            list: List IPv4 Addresses

            Example::

                [
                    "192.168.100.1",
                    "192.168.101.1"
                ]

            Returns empty list (``[]``) if absent

        """
        addresses = []
        # Filter L3 interfaces
        interfaces = [x for x in self.interface_lines if "l3" in x.flags]
        # Filter Standby Interfaces
        interfaces = [x for x in interfaces if x.standby is not None]
        # Filter VRF Interfaces
        if vrf == "global":
            interfaces = [x for x in interfaces if x.vrf is None]
        else:
            interfaces = [x for x in interfaces if x.vrf == vrf]

        for interface in interfaces:
            for group in interface.standby["groups"].values():
                addresses.extend([x["address"] for x in group["ipv4"]])
        return addresses






from ccutils.ccparser import BaseConfigLine
from ccutils.utils import CiscoRange
from ccutils.utils.common_utils import get_logger, split_interface_name
import re
import functools


class BaseInterfaceLine(BaseConfigLine):
    """
    Object for retrieving various config options on the interface level.
    """
    _ip_addr_regex = re.compile(pattern=r"^\sip\saddress\s(?P<ip_address>(?:\d{1,3}\.){3}\d{1,3})\s(?P<mask>(?:\d{1,3}\.){3}\d{1,3})(?:\s(?P<secondary>secondary))?", flags=re.MULTILINE)
    _description_regex = re.compile(pattern=r"^\sdescription\s(?P<description>.*)")
    _vrf_regex = re.compile(pattern=r"^(?:\sip)?\svrf\sforwarding\s(?P<vrf>\S+)", flags=re.MULTILINE)
    _shutdown_regex = re.compile(pattern=r"^\sshutdown", flags=re.MULTILINE)
    _cdp_regex = re.compile(pattern=r"cdp enable")
    _logging_event_regex = re.compile(pattern=r"^\slogging\sevent\s(?P<logging_event>\S+)", flags=re.MULTILINE)

    _standby_ip_regex = re.compile(pattern=r"^\sstandby\s(?P<standby_group>\d+)\sip\s(?P<ip_address>(?:\d{1,3}\.){3}\d{1,3})(?:\s(?P<secondary>secondary))?", flags=re.MULTILINE)
    _standby_timers_regex = re.compile(pattern=r"^\sstandby\s(?P<standby_group>\d+)\stimers\s(?P<hello>\d+)\s(?P<hold>\d+)", flags=re.MULTILINE)
    _standby_priority_regex = re.compile(pattern=r"^\sstandby\s(?P<standby_group>\d+)\spriority\s(?P<priority>\d+)", flags=re.MULTILINE)
    _standby_preempt_regex = re.compile(pattern=r"^\sstandby\s(?P<standby_group>\d+)\s(?P<preempt>preempt)", flags=re.MULTILINE)
    _standby_authentication_regex = re.compile(pattern=r"^\sstandby\s(?P<standby_group>\d+)\sauthentication\s(?P<authentication_type>md5)\skey-string\s(?P<key_type>0|7)\s(?P<key_string>\S+)", flags=re.MULTILINE)
    _standby_version_regex = re.compile(pattern=r"^\sstandby\sversion\s(?P<version>2)")
    
    _helper_address_regex = re.compile(pattern=r"^\sip\shelper-address\s(?P<helper_address>(?:\d{1,3}\.){3}\d{1,3})", flags=re.MULTILINE)


    _native_vlan_regex = re.compile(pattern=r"^ switchport trunk native vlan (?P<native_vlan>\d+)", flags=re.MULTILINE)
    _trunk_encapsulation_regex = re.compile(pattern=r"^ switchport trunk encapsulation (?P<encapsulation>dot1q|isl|negotiate)", flags=re.MULTILINE)
    _switchport_mode_regex = re.compile(pattern=r"^ switchport mode (?P<switchport_mode>access|trunk|dot1q-tunnel|private-vlan|dynamic)")
    _switchport_nonegotiate_regex = re.compile(pattern=r"^ switchport nonegotiate")
    _trunk_allowed_vlans_regex = re.compile(pattern=r"^ switchport trunk allowed vlan(?: add)? (?P<allowed_vlans>\S+)", flags=re.MULTILINE)
    _access_vlan_regex = re.compile(pattern=r"^ switchport access vlan (?P<access_vlan>\d+)", flags=re.MULTILINE)
    _voice_vlan_regex = re.compile(pattern=r"^ switchport voice vlan (?P<voice_vlan>\d+)")

    _channel_group_regex = re.compile(pattern=r"^ channel-group (?P<channel_group_number>\d+) mode (?P<channel_group_mode>\S+)")
    _speed_regex = re.compile(pattern=r"^ speed (?P<speed>\d+)", flags=re.MULTILINE)
    _duplex_regex = re.compile(pattern=r"^ duplex (?P<duplex>full|half)", flags=re.MULTILINE)
    _load_interval_regex = re.compile(pattern=r"^ load-interval (?P<load_interval>\d+)")

    _bandwidth_regex = re.compile(pattern=r"^ bandwidth (?P<bandwidth>\d+)", flags=re.MULTILINE)
    _delay_regex = re.compile(pattern=r"^ delay (?P<delay>\d+)", flags=re.MULTILINE)
    _mtu_regex = re.compile(pattern=r"^ mtu (?P<mtu>\d+)", flags=re.MULTILINE)
    _ip_mtu_regex = re.compile(pattern=r"^ ip mtu (?P<ip_mtu>\d+)", flags=re.MULTILINE)
    _ip_tcp_mss_regex = re.compile(pattern=r"^ ip tcp adjust-mss (?P<tcp_mss>\d+)", flags=re.MULTILINE)
    _keepalive_regex = re.compile(pattern=r"^ keepalive (?P<period>\d+) (?P<retries>\d+)")


    _service_policy_regex = re.compile(pattern=r"^ service-policy (?P<direction>input|output) (?P<policy_map>\S+)", flags=re.MULTILINE)
    _tunnel_src_regex = re.compile(pattern=r"^ tunnel source (?P<source>\S+)", flags=re.MULTILINE)
    _tunnel_dest_regex = re.compile(pattern=r"^ tunnel destination (?P<destination>\S+)", flags=re.MULTILINE)
    _tunnel_vrf_regex = re.compile(pattern=r"^ tunnel vrf (?P<vrf>\S+)", flags=re.MULTILINE)
    _tunnel_mode_regex = re.compile(pattern=r"^ tunnel mode (?P<mode>.*?)$", flags=re.MULTILINE)
    _tunnel_ipsec_profile_regex = re.compile(pattern=r"^ tunnel protection ipsec profile (?P<ipsec_profile>\S+)", flags=re.MULTILINE)

    _storm_control_threshold_regex = re.compile(pattern=r"^ storm-control (?P<traffic>broadcast|multicast) level (?:(?P<type>bps|pps) )?(?P<raising>\d{1,3}(?:\.\d{1,2})?(?:k|m|g)?)(?: (?P<falling>\d{1,3}(?:\.\d{1,2})?(?:k|m|g)?))?", flags=re.MULTILINE)
    _storm_control_action_regex = re.compile(pattern=r"^ storm-control action (?P<action>trap|shutdown)")
    _device_tracking_attach_policy_regex = re.compile(pattern=r"^ device-tracking attach-policy (?P<policy>\S+)")
    _encapsulation_regex = re.compile(pattern=r"^ encapsulation (?P<type>\S+) (?P<tag>\d+)(?: (?P<native>native))?")

    _service_instance_regex = re.compile(pattern=r"^ service instance (?P<number>\d+) (?P<type>\S+)")
    _service_instance_description_regex = re.compile(pattern=r"^  description (?P<description>.*)")
    _service_instance_encapsulation_mapping_regex = re.compile(pattern=r"^  encapsulation (?P<type>\S+) (?P<tag>\d+)$", flags=re.MULTILINE)
    _service_instance_encapsulation_string_regex = re.compile(pattern=r"^  encapsulation (?P<encapsulation>\S+)$", flags=re.MULTILINE)
    _service_instance_bridge_domain_regex = re.compile(pattern=r"^  bridge-domain (?P<number>\d+)$", flags=re.MULTILINE)
    _service_instance_service_policy_regex = re.compile(pattern=r"^  service-policy (?P<direction>input|output) (?P<policy_map>\S+)$", flags=re.MULTILINE)

    _ospf_process_regex = re.compile(pattern=r"^ ip ospf (?P<process_id>\d+) area (?P<area>\d+)$", flags=re.MULTILINE)
    _ospf_network_type_regex = re.compile(pattern=r"^ ip ospf network (?P<network_type>\S+)", flags=re.MULTILINE)
    _ospf_priority_regex = re.compile(pattern=r"^ ip ospf priority (?P<priority>\d+)", flags=re.MULTILINE)
    _ospf_cost_regex = re.compile(pattern=r"^ ip ospf cost (?P<cost>\d+)", flags=re.MULTILINE)




    def __init__(self, number, text, config, verbosity=3, name="BaseInterfaceLine"):
        """
        **This class is not meant to be instantiated directly, but only from BaseConfigParser instance.**

        Args:
            number (int): Index of line in config
            text (str): Text of the config line
            config (:obj:`BaseConfigParser`): Reference to the parent BaseConfigParser object
            verbosity (:obj:`int`, optional): Logging output level, defaults to 3: Warning

        """
        super(BaseInterfaceLine, self).__init__(number=number, text=text, config=config, verbosity=verbosity, name=name)

    @functools.lru_cache()
    def get_unprocessed(self, return_type=None):
        """
        Return a list of config lines under the interface, which did not match any of the existing regex patterns.
        Mostly for development/testing purposes.

        By default returns list of objects.

        Args:
            return_type (str): Set this to `"text"` to receive list of strings

        Returns:
            list: List of unprocessed config lines

        """
        unprocessed_children = self.get_children()
        regexes = [
            self._description_regex,
            self._ip_addr_regex,
            self._vrf_regex,
            self._shutdown_regex,
            self._ospf_priority_regex,
            self._ospf_process_regex,
            self._ospf_network_type_regex,
            self._ospf_cost_regex,
            self._cdp_regex,
            self._logging_event_regex,
            self._standby_ip_regex,
            self._standby_timers_regex,
            self._standby_priority_regex,
            self._standby_preempt_regex,
            self._standby_authentication_regex,
            self._standby_version_regex,
            self._helper_address_regex,
            self._native_vlan_regex,
            self._trunk_encapsulation_regex,
            self._switchport_mode_regex,
            self._switchport_nonegotiate_regex,
            self._trunk_allowed_vlans_regex,
            self._access_vlan_regex,
            self._voice_vlan_regex,
            self._channel_group_regex,
            self._speed_regex,
            self._duplex_regex,
            self._load_interval_regex,
            self._bandwidth_regex,
            self._delay_regex,
            self._mtu_regex,
            self._ip_mtu_regex,
            self._ip_tcp_mss_regex,
            self._keepalive_regex,
            self._service_policy_regex,
            self._tunnel_src_regex,
            self._tunnel_dest_regex,
            self._tunnel_vrf_regex,
            self._tunnel_mode_regex,
            self._tunnel_ipsec_profile_regex,
            self._storm_control_threshold_regex,
            self._storm_control_action_regex,
            self._encapsulation_regex,
            self._service_instance_regex,
            self._service_instance_description_regex,
            self._service_instance_encapsulation_mapping_regex,
            self._service_instance_encapsulation_string_regex,
            self._service_instance_bridge_domain_regex,
            self._service_instance_service_policy_regex,
            re.compile(pattern=r"^\sno\sip\saddress", flags=re.MULTILINE),
            re.compile(pattern=r"^ (no )?switchport$", flags=re.MULTILINE),
            re.compile(pattern=r"^ spanning-tree portfast")
        ]
        for regex in regexes:
            for child in self.re_search_children(regex=regex):
                unprocessed_children.remove(child)
        if return_type == "text":
            return [x.text for x in unprocessed_children]

    def _val_to_bool(self, entry, key):
        if entry[key]:
            entry[key] = True
        else:
            entry[key] = False
        return entry

    @property
    @functools.lru_cache()
    def flags(self):
        """
        List of flags/tags describing basic properties of the interface. Used for filtering purposes.
        Currently supported flags are:

        ``l2`` - Interface is switched port

        ``l3`` - Interface is routed port

        ``physical`` - Interface is a physical interface (Only \*Ethernet interfaces)

        ``svi`` - Interface is SVI (VLAN Interface)

        ``port-channel`` - Interface is port-channel

        ``pc-member`` - Interface is a member of Port-channel

        ``tunnel`` - Interface is a Tunnel

        Returns:
             list: List of flags

        """
        flags = []
        flags.append(self.port_mode)
        interface = split_interface_name(self.name)
        if "Vlan" in interface[0]:
            flags.append("svi")
        elif "Ethernet" in interface[0]:
            flags.append("physical")
            if self.channel_group is not None:
                flags.append("pc-member")
        elif "Port-channel" in interface[0]:
            flags.append("port-channel")
        elif "Tunnel" in interface[0]:
            flags.append("tunnel")
        elif "Loopback" in interface[0]:
            flags.append("virtual")
        elif "BDI" in interface[0]:
            flags.append("virtual")
        return flags

    @property
    @functools.lru_cache()
    def interface_name(self):
        self.logger.warning("DEPRECATED: You are using deprecated property .interface_name, use .name instead.")
        return self.name

    @property
    @functools.lru_cache()
    def name(self):
        """
        Return name of the interface, such as `GigabitEthernet0/1`.

        Returns:
            str: Name of the interface

        """
        if not self.is_interface:
            return None
        else:
            return self.re_match(self._interface_regex, group=1)

    @property
    @functools.lru_cache()
    def interface_description(self):
        self.logger.warning("DEPRECATED: You are using deprecated property .interface_description, use .description instead.")
        return self.description

    @property
    @functools.lru_cache()
    def description(self):
        """
        Returns description of the interface.

        Returns:
            str: Interface description

            Returns ``None`` if absent

        """
        if not self.is_interface:
            return None
        else:
            regex = r"description\s(.*)"
            candidates = self.re_search_children(regex=self._description_regex, group="description")
            if len(candidates) == 1:
                return candidates[0]
            else:
                return None

    @property
    @functools.lru_cache()
    def port_mode(self):
        """
        Checks whether the interface is running in switched (**l2**) or routed (**l3**) mode.

        Returns:
            str: `l2` or `l3`

        """
        if len(self.re_search_children(regex="ip address")):
            return "l3"
        else:
            return "l2"

    @property
    @functools.lru_cache()
    def ip_addresses(self):
        """
        Return list of IP addresses present on the interface

        Returns:
            list: List of dictionaries representing individual IP addresses

            Example::

                [
                    {
                        "ip_address": "10.0.0.1",
                        "mask": "255.255.255.0",
                        "secondary": False
                    },
                    {
                        "ip_address": "10.0.1.1",
                        "mask": "255.255.255.0",
                        "secondary": True
                    }
                ]

            If there is no IP address present on the interface, an empty list is returned.

        """
        ip_addresses = []
        candidates = self.re_search_children(regex=self._ip_addr_regex)
        for candidate in candidates:
            ip_addresses.append(self._val_to_bool(entry=candidate.re_search(regex=self._ip_addr_regex, group="ALL"), key="secondary"))
        return ip_addresses

    @property
    @functools.lru_cache()
    def vrf(self):
        """
        Return VRF of the interface

        Returns:
            str: Name of the VRF

            Returns ``None`` if absent

        """
        vrf = None
        candidates = self.re_search_children(regex=self._vrf_regex, group="vrf")
        if len(candidates):
            vrf = candidates[0]
        return vrf

    @property
    @functools.lru_cache()
    def shutdown(self):
        if len(self.re_search_children(regex=self._shutdown_regex)):
            return True
        else:
            return False

    @property
    @functools.lru_cache()
    def ospf_priority(self):
        """
        Returns OSPF priority of the interface.

        Returns:
            int: OSPF Priority or None

        """
        self.logger.warning("DEPRECATED: You are using deprecated property .ospf_priority, use .ospf['priority'] instead.")
        ospf_priority = None
        candidates = self.re_search_children(regex=self._ospf_priority_regex, group="priority")
        if len(candidates):
            ospf_priority = int(candidates[0])
        return ospf_priority

    @property
    @functools.lru_cache()
    def ospf(self):
        """
        Return OSPF interface parameters

        Returns:
            dict: OSPF parameters

            Example::

                {"process_id": 1, "area": 0, "network_type": "point-to-point", "priority": 200}

            Returns ``None`` if absent

        """
        ospf = None
        # OSPF Process Section
        candidates = self.re_search_children(regex=self._ospf_process_regex, group="ALL")
        if len(candidates):
            ospf = candidates[0]
            ospf["process_id"] = int(ospf["process_id"])
            ospf["area"] = int(ospf["area"])
        # OSPF Network Type Section
        candidates = self.re_search_children(regex=self._ospf_network_type_regex, group="ALL")
        if len(candidates):
            if ospf is None:
                ospf = {k: None for k in self._ospf_process_regex.groupindex.keys() if isinstance(k, str)}
            ospf.update(candidates[0])
        else:
            if ospf is not None:
                ospf.update({k: None for k in self._ospf_network_type_regex.groupindex.keys()})
        # OSPF Priority Section
        candidates = self.re_search_children(regex=self._ospf_priority_regex, group="ALL")
        if len(candidates):
            if ospf is None:
                ospf = {k: None for k in self._ospf_process_regex.groupindex.keys() if isinstance(k, str)}
                ospf.update({k: None for k in self._ospf_network_type_regex.groupindex.keys()})
            ospf.update(candidates[0])
            # Convert priority to int
            ospf["priority"] = int(ospf["priority"])
        else:
            if ospf is not None:
                ospf.update({k: None for k in self._ospf_priority_regex.groupindex.keys()})
        # OSPF Cost Section
        candidates = self.re_search_children(regex=self._ospf_cost_regex, group="ALL")
        if len(candidates):
            if ospf is None:
                ospf = {k: None for k in self._ospf_process_regex.groupindex.keys() if isinstance(k, str)}
                ospf.update({k: None for k in self._ospf_network_type_regex.groupindex.keys()})
                ospf.update({k: None for k in self._ospf_priority_regex.groupindex.keys()})
            ospf.update(candidates[0])
            # Convert cost to int
            ospf["cost"] = int(ospf["cost"])
        else:
            if ospf is not None:
                ospf.update({k: None for k in self._ospf_cost_regex.groupindex.keys()})


        return ospf

    @property
    @functools.lru_cache()
    def cdp(self):
        """
        Checks whether CDP is enabled on the interface. This property takes global CDP configuration into account,
        meaning if there is no specific configuration on the interface level, it will return state based on the entire
        config (eg. `no cdp run` in the global config will make this property be `False`)

        Returns:
            bool: ``True`` if CDP is enabled, ``False`` otherwise

        """
        global_cdp = self.config.cdp
        candidates = self.re_search_children(regex=self._cdp_regex)
        if len(candidates):
            if "no" in candidates[0].text:
                return False
            else:
                return True
        else:
            return global_cdp

    @property
    @functools.lru_cache()
    def logging_events(self):
        return self.re_search_children(regex=self._logging_event_regex, group="logging_event")

    @property
    @functools.lru_cache()
    def standby(self):
        """
        HSRP related configuration. Groups, IP addresses, hello/hold timers, priority and authentication.


        :return: Dictionary with top level keys being HSRP groups.
        """

        data = {"version": 1}
        if self.re_search_children(regex=self._standby_version_regex, group="version"):
            data["version"] = 2
        standby_ips = [self._val_to_bool(entry=x, key="secondary") for x in self.re_search_children(regex=self._standby_ip_regex, group="ALL")]
        #print(standby_ips)
        standby_timers = self.re_search_children(regex=self._standby_timers_regex, group="ALL")
        #print(standby_timers)
        standby_priority = self.re_search_children(regex=self._standby_priority_regex, group="ALL")
        #print(standby_priority)
        standby_preempt = [self._val_to_bool(entry=x, key="preempt") for x in self.re_search_children(regex=self._standby_preempt_regex, group="ALL")]
        #print(standby_preempt)
        standby_authentication = self.re_search_children(regex=self._standby_authentication_regex, group="ALL")
        #print(standby_authentication)
        if not len(standby_ips):
            return None
        data["groups"] = {}
        for entry in standby_ips:
            if entry["standby_group"] not in data["groups"].keys():
                data["groups"][entry["standby_group"]] = {"ip_addresses": []}
            data["groups"][entry["standby_group"]]["ip_addresses"].append(entry)
        for entry in standby_timers:
            data["groups"][entry["standby_group"]]["hello"] = entry["hello"]
            data["groups"][entry["standby_group"]]["hold"] = entry["hold"]
        for entry in standby_priority:
            data["groups"][entry["standby_group"]]["priority"] = entry["priority"]
        for entry in standby_preempt:
            data["groups"][entry["standby_group"]]["preempt"] = entry["preempt"]
        for entry in standby_authentication:
            data["groups"][entry["standby_group"]]["authentication_type"] = entry["authentication_type"]
            data["groups"][entry["standby_group"]]["key_type"] = entry["key_type"]
            data["groups"][entry["standby_group"]]["key_string"] = entry["key_string"]
        return data

    @property
    @functools.lru_cache()
    def helper_address(self):
        """
        Return a list of IP addresses specified with **ip helper-address** command (DHCP relay).

        Returns:
            list: List of helper addresses

            Returns ``None`` if absent

        """
        helper_address = None
        candidates = self.re_search_children(regex=self._helper_address_regex, group="helper_address")
        if len(candidates):
            helper_address = candidates
        return helper_address

    @property
    @functools.lru_cache()
    def native_vlan(self):
        """
        Return Native VLAN of L2 Interface

        Returns:
            int: Native VLAN Number (`None` if absent)

            Returns ``None`` if absent

        """
        native_vlan = None
        candidates = self.re_search_children(regex=self._native_vlan_regex, group="native_vlan")
        if len(candidates):
            native_vlan = int(candidates[0])
        return native_vlan

    @property
    @functools.lru_cache()
    def trunk_encapsulation(self):
        """
        Return encapsulation on trunk interfaces

        Returns:
            str: "dot1q" or "isl"

            Returns ``None`` if absent

        """
        trunk_encapsulation = None
        candidates = self.re_search_children(regex=self._trunk_encapsulation_regex, group="encapsulation")
        if len(candidates):
            trunk_encapsulation = candidates[0]
        return trunk_encapsulation

    @property
    @functools.lru_cache()
    def encapsulation(self):
        """
        Return encapsulation type and tag for subinterfaces

        Returns:
            dict: Encapsualtion parameters

            Example::

                {"type": "dot1q", "tag": 10, "native": False}

            Returns ``None`` if absent

        """
        encapsulation = None
        candidates = self.re_search_children(regex=self._encapsulation_regex, group="ALL")
        if len(candidates):
            encapsulation = candidates[0]
            encapsulation["tag"] = int(encapsulation["tag"])
            encapsulation["native"] = True if encapsulation["native"] == "native" else False
        return encapsulation

    @property
    @functools.lru_cache()
    def switchport_mode(self):
        """
        Return L2 Mode of interface, either access or trunk

        Returns:
            str: "access" or "trunk"

            Returns ``None`` if absent

        """
        switchport_mode = None
        candidates = self.re_search_children(regex=self._switchport_mode_regex, group="switchport_mode")
        if len(candidates):
            switchport_mode = candidates[0]
        return switchport_mode

    @property
    @functools.lru_cache()
    def switchport_nonegotiate(self):
        """
        Check whether the port is running DTP or not. Checks for presence of ``switchport nonegotiate`` command

        Returns:
            bool: ``True`` if command is present, ``False`` otherwise

        """
        candidates = self.re_search_children(regex=self._switchport_nonegotiate_regex)
        if len(candidates):
            return True
        else:
            return False

    @property
    @functools.lru_cache()
    def trunk_allowed_vlans(self):
        """
        Return a expanded list of VLANs allowed with ``switchport trunk allowed vlan x,y,z``.

        **Caution:** This does not mean the interface is necessarily a trunk port.

        Returns:
            list: Expanded list of allowed VLANs

            Returns ``None`` if absent

            Returns "none" if ``switchport trunk allowed vlan none``

        """
        candidates = self.re_search_children(regex=self._trunk_allowed_vlans_regex, group="allowed_vlans")
        if len(candidates):
            if len(candidates) == 1:
                # In case all VLANs are disabled - "switchport trunk allowed vlan none"
                if candidates[0] == "none":
                    return "none"

            self.logger.debug("Interface [{}]".format(self.name))
            candidates = ",".join(candidates)
            crange = CiscoRange(text=candidates)
            return crange._list
        else:
            return None

    @property
    @functools.lru_cache()
    def access_vlan(self):
        """
        Return a number of access VLAN or `None` if the command ``switchport access vlan x`` is not present.

        **Caution:** This does not mean the interface is necessarily an access port.

        Returns:
            int: Number of access VLAN or None

            Returns ``None`` if absent

        """
        access_vlan = None
        candidates = self.re_search_children(regex=self._access_vlan_regex, group="access_vlan")
        if len(candidates):
            access_vlan = int(candidates[0])
        return access_vlan

    @property
    @functools.lru_cache()
    def voice_vlan(self):
        """
        Return a number of voice VLAN

        Returns:
            int: Number of voice VLAN or None

            Returns ``None`` if absent

        """
        voice_vlan = None
        candidates = self.re_search_children(regex=self._voice_vlan_regex, group="voice_vlan")
        if len(candidates):
            voice_vlan = int(candidates[0])
        return voice_vlan

    @property
    @functools.lru_cache()
    def channel_group(self):
        """
        Return a dictionary describing Port-channel/Etherchannel related configuration

        Returns:
            dict: Channel-group parameters

            Example::

                {"channel_group_number": "1", "channel_group_mode": "active"}

            Otherwise returns ``None``

        """
        channel_group = None
        candidates = self.re_search_children(regex=self._channel_group_regex, group="ALL")
        if len(candidates):
            channel_group = candidates[0]
        return channel_group

    @property
    @functools.lru_cache()
    def speed(self):
        """
        Return speed of the interface set by command **speed X**

        Returns:
            int: Speed

            Returns ``None`` if absent

        """
        speed = None
        candidates = self.re_search_children(regex=self._speed_regex, group="speed")
        if len(candidates):
            speed = int(candidates[0])
        return speed

    @property
    @functools.lru_cache()
    def duplex(self):
        """
        Return duplex of the interface set by command **duplex X**.

        Returns:
            str: Duplex

            Returns ``None`` if absent

        """
        duplex = None
        candidates = self.re_search_children(regex=self._duplex_regex, group="duplex")
        if len(candidates):
            duplex = candidates[0]
        return duplex

    @property
    @functools.lru_cache()
    def bandwidth(self):
        """
        Return bandwidth of the interface set by command **bandwidth X**.

        Returns:
            int: Bandwidth

            Returns ``None`` if absent

        """
        bandwith = None
        candidates = self.re_search_children(regex=self._bandwidth_regex, group="bandwidth")
        if len(candidates):
            bandwith = candidates[0]
        return bandwith

    @property
    @functools.lru_cache()
    def delay(self):
        """
        Return delay of the interface set by command **delay X**.

        Returns:
            int: Delay

            Returns ``None`` if absent

        """
        delay = None
        candidates = self.re_search_children(regex=self._delay_regex, group="delay")
        if len(candidates):
            delay = candidates[0]
        return delay

    @property
    @functools.lru_cache()
    def mtu(self):
        """
        Return MTU of the interface set by command **mtu X**.

        Returns:
            int: MTU

            Returns ``None`` if absent

        """
        mtu = None
        candidates = self.re_search_children(regex=self._mtu_regex, group="mtu")
        if len(candidates):
            mtu = int(candidates[0])
        return mtu

    @property
    @functools.lru_cache()
    def ip_mtu(self):
        """
        Return IP MTU of the interface set by command **ip mtu X**.

        Returns:
            int: IP MTU

            Returns ``None`` if absent

        """
        ip_mtu = None
        candidates = self.re_search_children(regex=self._ip_mtu_regex, group="ip_mtu")
        if len(candidates):
            ip_mtu = int(candidates[0])
        return ip_mtu

    @property
    @functools.lru_cache()
    def tcp_mss(self):
        """
        Return TCP Max Segment Size of the interface set by command **ip tcp adjust-mss X**.

        Returns:
            int: TCP MSS

            Returns ``None`` if absent

        """
        tcp_mss = None
        candidates = self.re_search_children(regex=self._ip_tcp_mss_regex, group="tcp_mss")
        if len(candidates):
            tcp_mss = int(candidates[0])
        return tcp_mss

    @property
    @functools.lru_cache()
    def load_interval(self):
        """
        Return Load Interval of the interface set by command **load-interval X**.

        Returns:
            int: Load Interval

            Returns ``None`` if absent

        """
        load_interval = None
        candidates = self.re_search_children(regex=self._load_interval_regex, group="load_interval")
        if len(candidates):
            load_interval = int(candidates[0])
        return load_interval

    @property
    @functools.lru_cache()
    def keepalive(self):
        keepalive = None
        candidates = self.re_search_children(regex=self._keepalive_regex, group="ALL")
        if len(candidates):
            keepalive = {k: int(v) for k, v in candidates[0].items()}
        return keepalive

    @property
    @functools.lru_cache()
    def service_policy(self):
        """
        Return names of applied service policies

        Returns:
            dict: Dictionary containing names of both input and output policies.

            Example::

                {"input": "TEST_INPUT_POLICY", "output": "TEST_OUTPUT_POLICY"}

            If there are no policies specified, returns::

                {"input": None, "output": None}


        """
        service_policy = {"input": None, "output": None}
        candidates = self.re_search_children(regex=self._service_policy_regex, group="ALL")
        for candidate in candidates:
            if candidate["direction"] == "input":
                service_policy["input"] = candidate["policy_map"]
            elif candidate["direction"] == "output":
                service_policy["output"] = candidate["policy_map"]
        # print(candidates)
        return service_policy

    @property
    @functools.lru_cache()
    def service_instances(self):
        service_instances = None
        service_instance_candidates = self.re_search_children(regex=self._service_instance_regex)
        self.logger.debug("Interface {}: Service Instances Lines: {}".format(self.name, service_instance_candidates))
        if len(service_instance_candidates):
            service_instances = {}
        for service_instance_line in service_instance_candidates:
            match_dict = service_instance_line.re_search(regex=self._service_instance_regex, group="ALL")
            instance_number = int(match_dict["number"])
            service_instances[instance_number] = {"type": match_dict["type"]}

            # Description
            description_candidates = service_instance_line.re_search_children(regex=self._service_instance_description_regex, group="description")
            if len(description_candidates):
                service_instances[instance_number]["description"] = description_candidates[0]

            # Encapsulation
            # Try the mapping variant first
            encapsulation_candidates = service_instance_line.re_search_children(regex=self._service_instance_encapsulation_mapping_regex, group="ALL")
            if len(encapsulation_candidates):
                service_instances[instance_number]["encapsulation"] = {"type": encapsulation_candidates[0]["type"], "tag": int(encapsulation_candidates[0]["tag"])}
            else:
                # Fallback to string variant
                encapsulation_candidates = service_instance_line.re_search_children(regex=self._service_instance_encapsulation_string_regex, group="ALL")
                if len(encapsulation_candidates):
                    service_instances[instance_number]["encapsulation"] = encapsulation_candidates[0]["encapsulation"]

            # Bridge Domain
            bd_candidates = service_instance_line.re_search_children(regex=self._service_instance_bridge_domain_regex, group="number")
            if len(bd_candidates):
                service_instances[instance_number]["bridge_domain"] = int(bd_candidates[0])

            # Service Policy
            service_policy_candidates = service_instance_line.re_search_children(regex=self._service_instance_service_policy_regex, group="ALL")
            if len(service_policy_candidates):
                service_instances[instance_number]["service_policy"] = {"input": None, "output": None}
                for candidate in service_policy_candidates:
                    if candidate["direction"] == "input":
                        service_instances[instance_number]["service_policy"]["input"] = candidate["policy_map"]
                    elif candidate["direction"] == "output":
                        service_instances[instance_number]["service_policy"]["output"] = candidate["policy_map"]




        # print(service_instances)
        return service_instances

    @property
    @functools.lru_cache()
    def tunnel_properties(self):
        """
        Return properties related to Tunnel interfaces

        Returns:
            dict: Dictionary with tunnel properties.

            Example::

                {
                    "source": "Loopback0",
                    "destination": "10.0.0.1",
                    "vrf": None,
                    "mode": "ipsec ipv4",
                    "ipsec_profile": "TEST_IPSEC_PROFILE"
                }

            Returns ``None`` if absent

        """
        # Check if interface is a Tunnel
        if not re.match(pattern=r"^Tu", string=self.name):
            return None
        else:
            tunnel_properties = {}
            tunnel_src_candidates = self.re_search_children(regex=self._tunnel_src_regex, group="source")
            tunnel_dest_candidates = self.re_search_children(regex=self._tunnel_dest_regex, group="destination")
            tunnel_vrf_candidates = self.re_search_children(regex=self._tunnel_vrf_regex, group="vrf")
            tunnel_mode_candidates = self.re_search_children(regex=self._tunnel_mode_regex, group="mode")
            tunnel_ipsec_profile_candidates = self.re_search_children(regex=self._tunnel_ipsec_profile_regex, group="ipsec_profile")
            tunnel_properties["source"] = tunnel_src_candidates[0] if tunnel_src_candidates else None
            tunnel_properties["destination"] = tunnel_dest_candidates[0] if tunnel_dest_candidates else None
            tunnel_properties["vrf"] = tunnel_vrf_candidates[0] if tunnel_vrf_candidates else None
            tunnel_properties["mode"] = tunnel_mode_candidates[0] if tunnel_mode_candidates else None
            tunnel_properties["ipsec_profile"] = tunnel_ipsec_profile_candidates[0] if tunnel_ipsec_profile_candidates else None
            return tunnel_properties

    @property
    @functools.lru_cache()
    def storm_control(self):
        threshold_candidates = self.re_search_children(regex=self._storm_control_threshold_regex, group="ALL")
        action_candidates = self.re_search_children(regex=self._storm_control_action_regex, group="action")
        if len(threshold_candidates) == 0 and len(action_candidates) == 0:
            return None
        storm_control = {"thresholds": None, "action": None}
        if len(threshold_candidates):
            storm_control["thresholds"] = threshold_candidates
        if len(action_candidates):
            storm_control["action"] = action_candidates[0]
        return storm_control

    @property
    @functools.lru_cache()
    def device_tracking_policy(self):
        device_tracking_policy = None
        candidates = self.re_search_children(regex=self._device_tracking_attach_policy_regex, group="policy")
        if len(candidates):
            device_tracking_policy = candidates[0]
        return device_tracking_policy

    @property
    def ip_unnumbered_interface(self):
        self.logger.warning("This object does not support this property.")
        return None



from collections import MutableSequence
import re
import sys
from ccutils.utils.common_utils import get_logger
import json

class CiscoRange(MutableSequence):

    # TODO: Remove
    #PREFIX_REGEX = re.compile(pattern=r"^[A-z]{2,}")
    PREFIX_REGEX = re.compile(pattern=r"^[A-z\-]+(?=\d)", flags=re.MULTILINE)
    # SUFFIX_REGEX = re.compile(pattern=r"\d+(?:\/\d+)*(?:\s*-\s*\d+)*")
    SUFFIX_REGEX = re.compile(pattern=r"\d.*?$")
    RANGE_REGEX = re.compile(pattern=r"\d+\s*-\s*\d+")
    SUBINT_REGEX = re.compile(pattern=r"\.(?P<number>\d+)$")
    CHANNEL_REGEX = re.compile(pattern=r"\:(?P<number>\d+)")
    # SUFFIX_REGEX = re.compile(pattern=r"\d+(?:(?:-\d+)|(?:\/\d+)+(?:(?:\s*-\s*\d+)|(?:\.\d+(?:\s*-\s*\d+)))|(?:\.\d+))?$")
    # SUFFIX_REGEX = re.compile(pattern=r"(?P<prefix_slot>(?:\d+\/)*)(?:(?:\d+)|(?P<range>\d+\s*-\s*\d+)|(?:\d+\.(?P<subint>\d+)))$")

    # TODO: Remove
    #PREFIX_SLOT_REGEX = re.compile(pattern=r"(?P<prefix_slot>[A-z]{2,}(?:\d+/)*)(?P<number>\d+)")
    PREFIX_SLOT_REGEX = re.compile(pattern=r"(?P<prefix_slot>^[A-z\-]+(?=\d)(?:\d+/)*)(?P<number>\d+)", flags=re.MULTILINE)
    # PREFIX_SLOT_REGEX = re.compile(pattern=r"(?P<prefix_slot>^[A-z\-]+(?=\d)(?:\d+\/)*(?:\d+\.)?)(?P<number>\d+)", flags=re.MULTILINE)


    def __init__(self, text, verbosity=3):
        super(CiscoRange, self).__init__()
        self.logger = get_logger(name="CiscoRange", verbosity=verbosity)
        self.text = text
        self._list = self.sort_list(data=self.split_to_list(data=self.text))
        self.compressed_list = self.compress_list(data=self._list)

    def __delitem__(self, index):
        del self._list[index]
    
    def __getitem__(self, index):
        return self._list[index]
    
    def __len__(self):
        return len(self._list)
    
    def __setitem__(self, index, value):
        self._list[index] = value

    def __eq__(self, other):
        return self._list == other._list

    def __sub__(self, other):
        return self.sort_list(list(set(self._list) - set(other._list)))

    def __or__(self, other):
        return self.sort_list(list(set(self._list) | set(other._list)))

    def __xor__(self, other):
        return self.sort_list(list(set(self._list) ^ set(other._list)))

    def insert(self, index, value):
        self._list.insert(index, value)

    def __repr__(self):
        return "<CiscoRange: {}>".format(self.compressed_list)

    def add(self, data):
        data = self.split_to_list(data=data)
        if self.has_prefix(data=self._list):
            if not self.has_prefix(data=data):
                raise ValueError("Cannot merge prefixed and un-prefixed values!")
        else:
            if self.has_prefix(data=data):
                raise ValueError("Cannot merge prefixed and un-prefixed values!")
        _set = set(self._list)
        _set.add(tuple(data))
        self._list = self.sort_list(list(_set))
        self.compressed_list = self.compress_list(data=self._list)
    
    def remove(self, data):
        data = self.split_to_list(data=data)
        if self.has_prefix(data=self._list):
            if not self.has_prefix(data=data):
                raise ValueError("Cannot merge prefixed and un-prefixed values!")
        else:
            if self.has_prefix(data=data):
                raise ValueError("Cannot merge prefixed and un-prefixed values!")
        for element in data:
            try:
                self._list.remove(element)
            except Exception as e:
                pass
        self._list = self.sort_list(self._list)
        self.compressed_list = self.compress_list(data=self._list)


    def has_prefix(self, data):
        if isinstance(data, str):
            data = self.split_text(text=data)
        elif isinstance(data, list):
            pass
        else:
            self.logger.error(msg="Unexpected data type given: {}".format(type(data)))
            data = []
        # Check if at least one item contains prefix
        return bool(len(list(filter(lambda x: re.match(pattern=self.PREFIX_REGEX, string=x), data))))

    def check_prefix(self, data):
        if isinstance(data, str):
            data = self.split_text(text=data)
        elif isinstance(data, list):
            pass
        else:
            self.logger.error(msg="Unexpected data type given: {}".format(type(data)))
            data = []
        # Check if all items have prefix
        if len(data) == len(list(filter(lambda x: re.match(pattern=self.PREFIX_REGEX, string=x), data))):
            return True
        else:
            self.logger.error(msg="Not all items contain prefixes.")
            return False

    def split_to_list(self, data):
        _list = []
        raw_list = self.split_text(text=data)
        self.logger.debug("Raw List: {}".format(raw_list))
        if self.has_prefix(data=raw_list) and not self.check_prefix(data=data):
            self.logger.error(msg="Found prefix inconsistency in given data.")
            self.logger.debug("Returning '{}' for data: '{}'".format(_list, data))
            return _list
        for item in raw_list:
            results = self.split_item(item)
            for res in results:
                if res not in _list:
                    _list.append(res)
        self.logger.debug("Returning '{}' for data: '{}'".format(_list, data))
        return _list


    def split_text(self, text):
        result = []
        if isinstance(text, list):
            for item in text:
                if isinstance(item, int):
                    # TODO: Fix this ugliness
                    result.append(str(item))
                elif isinstance(item, str):
                    if item == "":
                        continue
                    else:
                        for subitem in [x.strip() for x in item.strip(",").split(",")]:
                            result.append(subitem)
            self.logger.debug("Text Instance: List - Returning {} for text: '{}'".format(result, text))
            return result
        elif isinstance(text, str):
            try:
                result = [x.strip() for x in text.strip(",").split(",")]
            except Exception as e:
                self.logger.error(msg="{}".format(repr(e)))
            finally:
                self.logger.debug("Text Instance: String - Returning {} for text: '{}'".format(result, text))
                return result
        else:
            self.logger.error(msg="Unexpected data type given: {}".format(type(text)))

    def split_item(self, item):
        self.logger.debug("Working on Item: {}".format(item))
        prefix_search = re.match(pattern=self.PREFIX_REGEX, string=item)
        prefix = prefix_search.group(0) if prefix_search else ""
        self.logger.debug(msg="Prefix: '{}'".format(prefix))
        suffix_search = self.SUFFIX_REGEX.search(string=item)
        if suffix_search:
            suffix = suffix_search.group(0)
            subint = self.SUBINT_REGEX.search(string=suffix).group("number") if self.SUBINT_REGEX.search(string=suffix) else None
            channel = self.CHANNEL_REGEX.search(string=suffix).group("number") if self.CHANNEL_REGEX.search(string=suffix) else None
            int_range = self.RANGE_REGEX.search(string=suffix).group(0) if self.RANGE_REGEX.search(string=suffix) else None
            if channel and subint:
                # Cut off channel and subint from suffix
                cutoff_length = (len(channel) + len(subint) + 2)
                suffix = suffix[:-cutoff_length]
            elif channel and (not subint):
                # Cut off channel from suffix
                cutoff_length = (len(channel) + 1)
                suffix = suffix[:-cutoff_length]
            elif (not channel) and subint:
                # Cut off subinterface from suffix, +1 for dot
                cutoff_length = (len(subint) + 1)
                suffix = suffix[:-cutoff_length]
            self.logger.debug("Prefix: {} Suffix: {} IntRange: {} Channel: {} SubInt: {}".format(prefix, suffix, int_range, channel, subint))

        else:
            self.logger.error("Suffix regex did not match on item: {}".format(item))
            suffix = ""
        self.logger.debug(msg="Suffix: '{}'".format(suffix))
        # Range and subinterface is not supported
        if int_range and (channel or subint):
            self.logger.error("Subinterfaces or channels cannot be combined with ranges. Item: {}".format(item))
            raise NotImplementedError("Subinterfaces or channels cannot be combined with ranges. Item: {}".format(item))
        elif int_range:
            # Range, no Subinterface or Channel
            base_suffix = suffix[:len(suffix) - len(int_range)]
            self.logger.debug(msg="Base Suffix: '{}'".format(base_suffix))
            start = int(int_range.split("-")[0].strip())
            stop = int(int_range.split("-")[1].strip())
            ERROR = "Given invalid range: '{}'. Start is bigger than stop!".format(item)
            assert stop >= start, ERROR
            number_range = range(start, stop+1)
            result = ["{}{}{}".format(prefix, base_suffix, x) for x in number_range]
            self.logger.debug("Result: {}".format(result))
            return result
        else:
            # Channel, Subinterface, no Range
            if channel and subint:
                result = ["{}{}:{}.{}".format(prefix, suffix, channel, subint)]
            # Channel, no Subinterface, no Range
            elif channel and (not subint):
                result = ["{}{}:{}".format(prefix, suffix, channel)]
            # No Channel, Subinterface, no Range
            elif (not channel) and subint:
                result = ["{}{}.{}".format(prefix, suffix, subint)]
            # No Channel, no Subinterface, no Range
            else:
                result = ["{}{}".format(prefix, suffix)]
            self.logger.debug("Result: {}".format(result))
            return result

    def sort_list(self, data):
        if self.has_prefix(data=data):
            results = []
            temp_list = []
            # Split item to list of prefixes and suffixes
            for item in data:
                try:
                    prefix = self.PREFIX_REGEX.search(string=item).group(0)
                except AttributeError as e:
                    self.logger.error("Item '{}' did not match PREFIX_REGEX".format(item))
                subint_search = self.SUBINT_REGEX.search(string=item)
                subint = subint_search.group("number") if subint_search else None
                channel_search = self.CHANNEL_REGEX.search(string=item)
                channel = channel_search.group("number") if channel_search else None
                suffix_list = []
                cutoff_length = 0
                if channel and subint:
                    # Cut off channel and subint from suffix
                    cutoff_length = (len(channel) + len(subint) + 2)
                elif channel and (not subint):
                    # Cut off channel from suffix
                    cutoff_length = (len(channel) + 1)
                elif (not channel) and subint:
                    # Cut off subinterface from suffix, +1 for dot
                    cutoff_length = (len(subint) + 1)

                if cutoff_length > 0:
                    suffix_list = [int(x) for x in self.SUFFIX_REGEX.search(string=item).group(0)[:-cutoff_length].split("/")]
                else:
                    suffix_list = [int(x) for x in self.SUFFIX_REGEX.search(string=item).group(0).split("/")]
                temp_entry = {
                    "item": item,
                    "prefix": prefix,
                    "suffix_list": suffix_list,
                    "channel": self.int_or_none(channel),
                    "subint": self.int_or_none(subint)
                }
                temp_list.append(temp_entry)

            self.logger.debug(msg="TempList: {}".format(temp_list))
            suffix_lengths = list(set([len(x["suffix_list"]) for x in temp_list]))
            suffix_lengths.sort()
            # Get items with same suffix length
            for suffix_length in suffix_lengths:
                items = list(filter(lambda x: len(x["suffix_list"]) == suffix_length, temp_list))
                self.logger.debug(msg="SuffixLength: {}: {}".format(suffix_length, items))
                # Assign numeric values to items based on suffixes
                for item in items:
                    number = 0
                    reverse_suffix = list(item["suffix_list"])
                    reverse_suffix.reverse()
                    for index, value in enumerate(reverse_suffix):
                        number += int(100*value) * 100**(index+2)
                    if item["channel"] is not None:
                        number += item["channel"] * 10e4
                    if item["subint"] is not None:
                        number += item["subint"]
                    item["number"] = number
                    self.logger.debug("Prefix: {} SuffixList: {}, Channel: {}, SubInt: {}, Value: {} ".format(
                        item["prefix"], item["suffix_list"], item["channel"], item["subint"], item["number"]))
                sorted_items = list(items)
                sorted_items.sort(key=lambda x: x["number"])
                for item in sorted_items:
                    results.append(item["item"])
                self.logger.debug(msg="SortedItems: {}".format(sorted_items))

            self.logger.debug(msg="Results: {}".format(results))
            return results        
        else:
            temp_list = [int(x) for x in data]
            temp_list.sort()
            
            return ["{}".format(x) for x in temp_list]

    def compress_list(self, data):
        results = []
        prefix = ""
        entry = []
        for index, item in enumerate(data):
            current_prefix = ""
            match = re.match(pattern=self.PREFIX_SLOT_REGEX, string=item)
            if match:
                current_prefix = match.group("prefix_slot")
                subint_search = self.SUBINT_REGEX.search(string=item)
                subint = subint_search.group("number") if subint_search else None
                channel_search = self.CHANNEL_REGEX.search(string=item)
                channel = channel_search.group("number") if channel_search else None
                if (subint is not None) or (channel is not None):
                    results.append(item)
                    continue
                item = int(match.group("number"))
            else:
                item = int(item)
            if not len(entry):
                self.logger.debug(msg="BEFORE: CurrentPrefix: '{}'; CurrentItem: '{}'; CurrentEntry: '{}'; Prefix: '{}'".format(current_prefix, item, entry, prefix))
                entry.append(item)
                prefix = current_prefix
            elif len(entry) == 1:
                self.logger.debug(msg="BEFORE: CurrentPrefix: '{}'; CurrentItem: '{}'; CurrentEntry: '{}'; Prefix: '{}'".format(current_prefix, item, entry, prefix))
                if current_prefix == prefix:
                    if entry[0] == item-1:
                        entry.append(item)
                    else:
                        results.append("{}{}".format(prefix, *entry))
                        entry = [item]
                else:
                    results.append("{}{}".format(prefix, *entry))
                    entry = [item]
                    prefix = current_prefix
            elif len(entry) == 2:
                self.logger.debug(msg="BEFORE: CurrentPrefix: '{}'; CurrentItem: '{}'; CurrentEntry: '{}'; Prefix: '{}'".format(current_prefix, item, entry, prefix))
                if current_prefix == prefix:
                    if entry[1] == item-1:
                        # Increase entry[1]
                        entry[1] += 1
                    else:
                        # Cisco VLAN range uses compression (eg. "1-3") only when the difference between start and stop is HIGHER than 1
                        # If difference is EQUAL to 1, each number is separate (eg. "1,2")
                        #Do this only for entries without prefix
                        if (entry[1] == entry[0] + 1) and prefix == "":
                            results.append("{}".format(entry[0]))
                            results.append("{}".format(entry[1]))
                        else:
                            results.append("{}{}-{}".format(prefix, *entry))
                        entry = [item]
                else:
                    results.append("{}{}-{}".format(prefix, *entry))
                    entry = [item]
                    prefix = current_prefix
        
        if len(entry) == 1:
            results.append("{}{}".format(prefix, *entry))
        elif len(entry) == 2:
            if (entry[1] == entry[0] + 1) and prefix == "":
                results.append("{}".format(entry[0]))
                results.append("{}".format(entry[1]))
            else:
                results.append("{}{}-{}".format(prefix, *entry))
        
        return results

    @staticmethod
    def int_or_none(item):
        if item is None:
            return None
        try:
            item = int(item)
        except Exception as e:
            item = None
        finally:
            return item

    def to_string(self):
        return ",".join(self.compressed_list)



def main():
    text = "1,3,4-6,8,11-20, 9"
    text2 = "Fa0, Fa0/3-6, Fa0/1-2, Fa2/1-2 Fa2/0/10-11, Fa1/0/3-4"
    text4 = "Fa0, Fa0/3-6, Fa0/1-2, Fa2/3-4 Fa2/0/10-11, Fa1/0/3-4"
    text3 = "1,2,10,15,19,21,22,24,27,30,31,55,58,66,100-102,"
    crange = CiscoRange(text=text)
    
    crange.add("22-24")
    print(crange)


if __name__ == "__main__":
     main()